"""The `AsyncJOptional` class is like the `JOptional` one but simplify the use of Awaitable values"""

# Copyright 2022 the original author or authors (i.e.: St4rG00se for Dev4py).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from functools import partial
from typing import Generic, Final, Optional, TypeAlias, cast, Any, Awaitable

from dev4py.utils.JOptional import JOptional
from dev4py.utils.awaitables import is_awaitable
from dev4py.utils.objects import async_require_non_none, non_none, is_none, require_non_none, to_self, to_none
from dev4py.utils.types import T, V, SyncOrAsync, Supplier, Function, R, Consumer, Runnable, Predicate

_VType: TypeAlias = SyncOrAsync[Optional[V]]  # pragma: no mutate


class AsyncJOptional(Generic[T]):
    """A JOptional class for async value mapping"""

    __CREATE_KEY: Final[object] = object()
    __NO_VALUE_ERROR_MSG: Final[str] = "No value present"
    __EMPTY: AsyncJOptional[Any]

    @classmethod
    def of(cls, value: _VType[T]) -> AsyncJOptional[T]:
        """
        Returns a JOptional describing the given non-None value or Awaitable value

        Args:
            value: The value (or Awaitable value) to describe, which must be non-None and not an Awaitable of None

        Returns:
            JOptional[T]: A JOptional of value T or Awaitable[T] type with the value present

        Raises:
            TypeError: Raises a TypeError on terminal operation if value is None or is an Awaitable of None
        """
        return AsyncJOptional(async_require_non_none(value, cls.__NO_VALUE_ERROR_MSG), cls.__CREATE_KEY)

    @classmethod
    def of_noneable(cls, value: _VType[T]) -> AsyncJOptional[T]:
        """
        Returns a JOptional describing the given value or Awaitable value, if non-None, otherwise returns an empty
        AsyncJOptional

        Args:
            value: The possibly-None or Awaitable of None value to describe

        Returns:
            JOptional[T]: A JOptional of value T  or Awaitable[T] type with the value present if the specified value is
            non-None, otherwise an empty AsyncJOptional
        """
        return AsyncJOptional(value, cls.__CREATE_KEY)

    @classmethod
    def empty(cls) -> AsyncJOptional[T]:
        """Return an empty instance of AsyncJOptional"""
        return cls.__EMPTY

    def __init__(self, value: _VType[T], create_key: object):
        """AsyncJOptional private constructor: Constructs an instance with the described value"""
        assert create_key == self.__CREATE_KEY, \
            "AsyncJOptional private constructor! Please use AsyncJOptional.of or AsyncJOptional.of_noneable"
        self._value: _VType[T] = value

    async def _get_value(self) -> Optional[T]:
        """
        Private method in order to manage Sync or Awaitable value

        Returns:
            Optional[T]: the current value
        """
        if is_awaitable(self._value):
            self._value = await cast(Awaitable[Optional[T]], self._value)
        return cast(Optional[T], self._value)

    async def is_present(self) -> bool:
        """
        If a value is present, returns true, otherwise false

        Note: A value is present when not None and not an Awaitable of None

        Returns:
            bool: true if a value is not None and not an Awaitable of None, otherwise false
        """
        return non_none(await self._get_value())

    async def is_empty(self) -> bool:
        """
        If a value is not present, returns true, otherwise false.

        Note: A value is empty when None or an Awaitable of None

        Returns:
            bool: true if a value is None or an Awaitable of None, otherwise false
        """
        return is_none(await self._get_value())

    def get(self) -> Awaitable[T]:
        """
        If a value is present, returns the value, otherwise raises a ValueError

        Returns:
            value: The non-None value described by this AsyncJOptional

        Raises:
            ValueError: Raises a ValueError if the value is empty
        """
        return self.or_else_raise()

    def or_else(self, other: Optional[T] = None) -> Awaitable[Optional[T]]:
        """
        If a value is present, returns the value, otherwise returns other

        Args:
            other: The value to be returned, if no value is present. May be None.

        Returns:
            value: The value, if present, otherwise other
        """
        return self.or_else_get(cast(Supplier[Optional[T]], partial(to_self, obj=other)))

    async def or_else_get(self, supplier: Supplier[Optional[T]] = to_none) -> Optional[T]:
        """
        If a value is present, returns the value, otherwise returns the result produced by the supplying function

        Args:
            supplier: The supplying function that produces a value to be returned

        Returns:
            value: The value, if present, otherwise the result produced by the supplying function

        Raises:
            TypeError: if the supplying function is None
        """
        require_non_none(supplier)
        return cast(Optional[T], self._value if await self.is_present() else supplier())

    @staticmethod  # pragma: no mutate
    def __or_else_raise_supplier_lambda() -> Exception:
        """
        private static method to replace or_else_raise lambda supplier by a function
        Note: lambda are not used in order to be compatible with multiprocessing (lambda are not serializable)
        """
        # lambda: cast(Exception, ValueError(AsyncJOptional.__NO_VALUE_ERROR_MSG))
        return cast(Exception, ValueError(AsyncJOptional.__NO_VALUE_ERROR_MSG))

    async def or_else_raise(
            self,
            supplier: Supplier[Exception] = __or_else_raise_supplier_lambda
    ) -> T:
        """
        If a value is present, returns the value, otherwise raises an exception produced by the exception supplying
        function

        Args:
            supplier: The supplying function that produces an exception to be raised (default: ValueError)

        Returns:
            value: The value, if present

        Raises:
            Exception: If value is empty
            TypeError: if the exception supplying function is None
        """
        require_non_none(supplier)
        if await self.is_empty():
            raise supplier()
        return cast(T, self._value)

    def or_get(self, supplier: Supplier[AsyncJOptional[T]]) -> AsyncJOptional[T]:
        """
        This is the `or` equivalent in java Optional (Reminder: `or` is a python keyword)

        If a value is present, returns an AsyncJOptional describing the value, otherwise returns an AsyncJOptional
        produced by the supplying function

        Args:
            supplier: The supplying function that produces an AsyncJOptional to be returned

        Returns:
            AsyncJOptional[T]: Returns an AsyncJOptional describing the value of this AsyncJOptional, if a value is
            present, otherwise an AsyncJOptional produced by the supplying function

        Raises:
            TypeError: if the supplying function is None
        """
        require_non_none(supplier)

        async def _async_or_get(current: AsyncJOptional[T], suppl: Supplier[AsyncJOptional[T]]) -> Optional[T]:
            return await current.get() if await current.is_present() else await require_non_none(suppl()).or_else_get()

        return AsyncJOptional.of_noneable(_async_or_get(self, supplier))

    @staticmethod  # pragma: no mutate
    def __map_lambda(v: T, mapper: Function[T, _VType[R]]) -> AsyncJOptional[R]:
        """
        private static method to replace inner map lambda supplier by a function
        Note: lambda are not used in order to be compatible with multiprocessing (lambda are not serializable)
        """
        # lambda v: AsyncJOptional.of_noneable(mapper(v))
        return AsyncJOptional.of_noneable(mapper(v))

    def map(self, mapper: Function[T, _VType[R]]) -> AsyncJOptional[R]:
        """
        If a value is present, returns an AsyncJOptional describing (as if by of_noneable) the result of applying the
        given mapping function to the value, otherwise returns an empty AsyncJOptional

        If the mapping function returns a None result then this method returns an empty AsyncJOptional.

        Note: mapping function can be sync or Awaitable (like async coroutine)

        Args:
            mapper: The mapping function to apply to a value, if present

        Returns:
            AsyncJOptional[R]: An AsyncJOptional describing the result of applying a mapping function to the value of
            this AsyncJOptional, if a value is present, otherwise an empty AsyncJOptional

        Raises:
            TypeError: If the mapping function is None
        """
        require_non_none(mapper)
        return self.flat_map(partial(AsyncJOptional.__map_lambda, mapper=mapper))

    def flat_map(self, mapper: Function[T, AsyncJOptional[R]]) -> AsyncJOptional[R]:
        """
        If a value is present, returns the result of applying the given AsyncJOptional-bearing mapping function to the
        value, otherwise returns an empty AsyncJOptional

        This method is similar to map(Function), but the mapping function is one whose result is already an
        AsyncJOptional, and if invoked, flatMap does not wrap it within an additional AsyncJOptional

        Args:
            mapper: The mapping function to apply to a value, if present

        Returns:
            AsyncJOptional[R]: The result of applying a AsyncJOptional-bearing mapping function to the value of this
            AsyncJOptional, if a value is present, otherwise an empty AsyncJOptional

        Raises:
            TypeError: if the mapping function is None
        """
        require_non_none(mapper)

        async def _async_flat_map(
                current: AsyncJOptional[T], flat_mapper: Function[T, AsyncJOptional[R]]
        ) -> Optional[R]:
            if await current.is_present():
                new_async_joptional: AsyncJOptional[R] = require_non_none(flat_mapper(await current.get()))
                return await new_async_joptional.or_else_get()
            return None

        return AsyncJOptional.of_noneable(_async_flat_map(self, mapper))

    async def if_present(self, consumer: Consumer[T]) -> None:
        """
        If a value is present, performs the given consumer with the value, otherwise does nothing

        Args:
            consumer: The consumer to be performed, if a value is present

        Returns:
            Nothing

        Raises:
            TypeError: if the consumer is None
        """
        require_non_none(consumer)
        # pylint: disable=W0106
        (await self.is_present()) and consumer(cast(T, self._value))

    async def if_empty(self, empty_action: Runnable) -> None:
        """
        If a value is not present, performs the given runnable, otherwise does nothing

        Args:
            empty_action: The runnable to be performed, if a value is not present

        Returns:
            Nothing

        Raises:
            TypeError: if the runnable is None
        """
        require_non_none(empty_action)
        # pylint: disable=W0106
        (await self.is_empty()) and empty_action()

    async def if_present_or_else(self, consumer: Consumer[T], empty_action: Runnable) -> None:
        """
        If a value is present, performs the given consumer with the value, otherwise performs the given empty_action

        Args:
            consumer: The consumer to be performed, if a value is present
            empty_action: The runnable to be performed, if a value is not

        Returns:
            Nothing

        Raises:
            TypeError: if the consumer is None or if the runnable is None
        """
        require_non_none(consumer)
        require_non_none(empty_action)
        # pylint: disable=W0106
        consumer(cast(T, self._value)) if await self.is_present() else empty_action()

    def filter(self, predicate: Predicate[T]) -> AsyncJOptional[T]:
        """
        If a value is present, and the value matches the given predicate, returns an AsyncJOptional describing the
        value, otherwise returns an empty AsyncJOptional

        Args:
            predicate: The predicate to apply to a value, if present

        Returns:
            AsyncJOptional[T]: An AsyncJOptional describing the value of this AsyncJOptional, if a value is present and
            the value matches the given predicate, otherwise an empty AsyncJOptional

        Raises:
            TypeError: if the predicate is None
        """
        require_non_none(predicate)

        async def _async_filter(current: AsyncJOptional[T]) -> Optional[T]:
            return await current.or_else_get() if await current.is_empty() or predicate(await current.get()) else None

        return AsyncJOptional.of_noneable(_async_filter(self))

    def peek(self, consumer: Consumer[T]) -> AsyncJOptional[T]:
        """
        If a value is present, performs the given consumer with the value, otherwise does nothing and returns the
        current equivalent of AsyncJOptional

        Args:
            consumer: The consumer to be performed, if a value is present

        Returns:
             AsyncJOptional[T]: The current equivalent of AsyncJOptional

        Raises:
            TypeError: if the consumer is None
        """
        require_non_none(consumer)

        async def _async_peek(current: AsyncJOptional[T], cons: Consumer[T]) -> Optional[T]:
            await current.if_present(cons)
            return await current.or_else_get()

        return AsyncJOptional.of_noneable(_async_peek(self, consumer))

    def to_joptional(self) -> JOptional[Awaitable[Optional[T]]]:
        """
        Convert the current AsyncJOptional to a JOptional

        Note: the returned JOptional will contain an Awaitable[Optional[T]] value

        Returns:
            JOptional[Awaitable[Optional[T]]]: The corresponding JOptional
        """
        return JOptional.of_noneable(cast(Awaitable[Optional[T]], self._get_value()))


# INIT STATIC VARIABLES
# noinspection PyProtectedMember
# noinspection PyUnresolvedReferences
# pylint: disable=W0212
AsyncJOptional._AsyncJOptional__EMPTY = AsyncJOptional(None, AsyncJOptional._AsyncJOptional__CREATE_KEY)  # type: ignore



"""lists module tests"""

# Copyright 2022 the original author or authors (i.e.: St4rG00se for Dev4py).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from pytest import raises

from dev4py.utils import lists


class TestEmptyList:
    """empty_list function tests"""

    class TestNominalCase:
        def test_should__return_an_empty_list(self) -> None:
            """Should return an empty list"""
            # GIVEN / WHEN
            result: list[int] = lists.empty_list()

            # THEN
            assert result == []


class TestAppend:
    """append function tests"""

    class TestNominalCase:
        def test_existing_parameters__should__return_the_list_with_added_value(self) -> None:
            """When list and value are set, should return the list with added value"""
            # GIVEN
            lst: list[int] = [1, 2, 3]
            value: int = 4

            # WHEN
            result: list[int] = lists.append(lst, value)

            # THEN
            expected_result: list[int] = [1, 2, 3, 4]
            assert result == expected_result
            assert lst == expected_result

        def test_none_value__should__return_the_list_with_added_value(self) -> None:
            """When list is set but value is none, should return the list with added value"""
            # GIVEN
            lst: list[Optional[int]] = [1, 2, 3]
            value: Optional[int] = None

            # WHEN
            result: list[Optional[int]] = lists.append(lst, value)

            # THEN
            expected_result: list[Optional[int]] = [1, 2, 3, None]
            assert result == expected_result
            assert lst == expected_result

    class TestErrorCase:
        def test_none_list__should__raise_type_error(self) -> None:
            """When the list is None, should raise TypeError exception"""
            # GIVEN
            value: int = 4

            # WHEN / THEN
            with raises(TypeError):
                # noinspection PyTypeChecker
                lists.append(None, value)


class TestExtend:
    """extend function tests"""

    class TestNominalCase:
        def test_existing_parameters__should__add_elements_to_the_first_list_and_return_it(self) -> None:
            """When the lists are set, should add all value from the second list to the first one and return it"""
            # GIVEN
            lst_1: list[int] = [1, 2, 3]
            lst_2: list[int] = [4, 5]

            # WHEN
            result: list[int] = lists.extend(lst_1, lst_2)

            # THEN
            expected_result: list[int] = [1, 2, 3, 4, 5]
            assert result == expected_result
            assert lst_1 == expected_result
            assert lst_2 == [4, 5]

    class TestErrorCase:
        def test_none_first_list__should__raise_type_error(self) -> None:
            """When the first list is None, should raise TypeError exception"""
            # GIVEN
            lst: list[int] = lists.empty_list()

            # WHEN / THEN
            with raises(TypeError):
                # noinspection PyTypeChecker
                lists.extend(None, lst)

        def test_none_second_list__should__raise_type_error(self) -> None:
            """When the second list is None, should raise TypeError exception"""
            # GIVEN
            lst: list[int] = lists.empty_list()

            # WHEN / THEN
            with raises(TypeError):
                # noinspection PyTypeChecker
                lists.extend(lst, None)



"""retry module tests"""

# Copyright 2022 the original author or authors (i.e.: St4rG00se for Dev4py).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from dataclasses import FrozenInstanceError
from time import time
from typing import Awaitable
from unittest.mock import patch, MagicMock

from pytest import raises

from dev4py.utils import retry
from dev4py.utils.retry import RetryConfiguration
from dev4py.utils.types import Function


##############################
#     PUBLIC MODULE PART     #
##############################
class TestRetryConfiguration:
    """RetryConfiguration class tests"""

    class TestConstructor:
        """Constructor tests"""

        class TestNominalCase:
            def test_existing_parameters__should__create_new_retry_configuration(self) -> None:
                """When all parameters are set, should create a new RetryConfiguration with given parameters"""
                # GIVEN
                exponent: int = 7
                delay: float = 0.5
                max_tries: int = 10

                # WHEN
                retry_config: RetryConfiguration = RetryConfiguration(
                    exponent=exponent, delay=delay, max_tries=max_tries
                )

                # THEN
                assert retry_config.exponent == exponent
                assert retry_config.delay == delay
                assert retry_config.max_tries == max_tries

            def test_min_valued_parameters__should__create_new_retry_configuration(self) -> None:
                """
                When all parameters are set to their min value, should create a new RetryConfiguration with given
                parameters
                """
                # GIVEN
                exponent: int = 0
                delay: float = 0
                max_tries: int = 1

                # WHEN
                retry_config: RetryConfiguration = RetryConfiguration(
                    exponent=exponent, delay=delay, max_tries=max_tries
                )

                # THEN
                assert retry_config.exponent == exponent
                assert retry_config.delay == delay
                assert retry_config.max_tries == max_tries

            def test_no_parameters__should__create_new_default_retry_configuration(self) -> None:
                """When no parameters are set, should create a new RetryConfiguration with default parameters"""
                # GIVEN / WHEN
                retry_config: RetryConfiguration = RetryConfiguration()

                # THEN
                assert retry_config.exponent == 2
                assert retry_config.delay == 0.1
                assert retry_config.max_tries == 3

        class TestErrorCase:
            def test_none_exponent__should_raise_type_error(self) -> None:
                """When exponent is None, should raise TypeError exception"""
                # GIVEN
                delay: float = 0.5
                max_tries: int = 10

                # WHEN / THEN
                with raises(TypeError) as error:
                    # noinspection PyTypeChecker
                    RetryConfiguration(exponent=None, delay=delay, max_tries=max_tries)

                assert str(error.value) == "exponent must be non None"

            def test_none_delay__should_raise_type_error(self) -> None:
                """When delay is None, should raise TypeError exception"""
                # GIVEN
                exponent: int = 7
                max_tries: int = 10

                # WHEN / THEN
                with raises(TypeError) as error:
                    # noinspection PyTypeChecker
                    RetryConfiguration(exponent=exponent, delay=None, max_tries=max_tries)

                assert str(error.value) == "delay must be non None"

            def test_none_max_tries__should_raise_type_error(self) -> None:
                """When max_tries is None, should raise TypeError exception"""
                # GIVEN
                exponent: int = 7
                delay: float = 0.5

                # WHEN / THEN
                with raises(TypeError) as error:
                    # noinspection PyTypeChecker
                    RetryConfiguration(exponent=exponent, delay=delay, max_tries=None)

                assert str(error.value) == "max_tries must be non None"

            def test_less_than_zero_exponent__should_raise_value_error(self) -> None:
                """When exponent is less than zero, should raise ValueError exception"""
                # GIVEN
                exponent: int = -1
                delay: float = 0.5
                max_tries: int = 3

                # WHEN / THEN
                with raises(ValueError) as error:
                    RetryConfiguration(exponent=exponent, delay=delay, max_tries=max_tries)

                assert str(error.value) == "exponent must be greater than or equals to 0"

            def test_less_than_zero_delay__should_raise_value_error(self) -> None:
                """When delay is less than zero, should raise ValueError exception"""
                # GIVEN
                exponent: int = 1
                delay: float = -0.5
                max_tries: int = 3

                # WHEN / THEN
                with raises(ValueError) as error:
                    RetryConfiguration(exponent=exponent, delay=delay, max_tries=max_tries)

                assert str(error.value) == "delay must be greater than or equals to 0"

            def test_less_than_one_max_tries__should_raise_value_error(self) -> None:
                """When max_tries is less than one, should raise ValueError exception"""
                # GIVEN
                exponent: int = 7
                delay: float = 0.5
                max_tries: int = 0

                # WHEN / THEN
                with raises(ValueError) as error:
                    RetryConfiguration(exponent=exponent, delay=delay, max_tries=max_tries)

                assert str(error.value) == "max_tries must be greater than or equals to 1"

    class TestSetters:
        """Setters tests"""

        class TestErrorCase:
            def test_set_new_exponent__should__raise_frozen_instance_error(self) -> None:
                """When set a new exponent, should raise a FrozenInstanceError exception"""
                # GIVEN
                exponent: int = 7
                delay: float = 0.5
                max_tries: int = 10
                retry_config: RetryConfiguration = RetryConfiguration(
                    exponent=exponent, delay=delay, max_tries=max_tries
                )

                # WHEN / THEN
                new_exponent: int = 8
                with raises(FrozenInstanceError):
                    # noinspection PyDataclass
                    retry_config.exponents = new_exponent

            def test_set_new_delay__should__raise_frozen_instance_error(self) -> None:
                """When set a new delay, should raise a FrozenInstanceError exception"""
                # GIVEN
                exponent: int = 7
                delay: float = 0.5
                max_tries: int = 10
                retry_config: RetryConfiguration = RetryConfiguration(
                    exponent=exponent, delay=delay, max_tries=max_tries
                )

                # WHEN / THEN
                new_delay: float = 2.1
                with raises(FrozenInstanceError):
                    # noinspection PyDataclass
                    retry_config.delay = new_delay

            def test_set_new_max_tries__should__raise_frozen_instance_error(self) -> None:
                """When set a new max_tries, should raise a FrozenInstanceError exception"""
                # GIVEN
                exponent: int = 7
                delay: float = 0.5
                max_tries: int = 10
                retry_config: RetryConfiguration = RetryConfiguration(
                    exponent=exponent, delay=delay, max_tries=max_tries
                )

                # WHEN / THEN
                new_max_tries: int = 5
                with raises(FrozenInstanceError):
                    # noinspection PyDataclass
                    retry_config.max_tries = new_max_tries

    class TestGetWaitingInterval:
        """get_waiting_interval method tests"""

        class TestNominalCase:
            def test_valid_retry_number__should__return_exponential_backoff_result(self) -> None:
                """When retry_number is valid, should return the exponential backoff depending on it"""
                # GIVEN
                exponent: int = 3
                delay: float = 0.5
                max_tries: int = 10
                retry_config: RetryConfiguration = RetryConfiguration(
                    exponent=exponent, delay=delay, max_tries=max_tries
                )
                retry_number: int = 5

                # WHEN
                backoff: float = retry_config.get_waiting_interval(retry_number=retry_number)

                # THEN
                assert backoff == delay * (exponent ** retry_number)

            def test_retry_number_equals_to_max_tries__should__return_exponential_backoff_result(self) -> None:
                """When retry_number is equals to max tries, should return the exponential backoff depending on it"""
                # GIVEN
                exponent: int = 3
                delay: float = 0.5
                max_tries: int = 10
                retry_config: RetryConfiguration = RetryConfiguration(
                    exponent=exponent, delay=delay, max_tries=max_tries
                )

                # WHEN
                backoff: float = retry_config.get_waiting_interval(retry_number=max_tries)

                # THEN
                assert backoff == delay * (exponent ** max_tries)

        class TestErrorCase:
            def test_none_retry_number__should__raise_type_error(self) -> None:
                """When retry_number is None, should raise a TypeError exception"""
                # GIVEN
                exponent: int = 7
                delay: float = 0.5
                max_tries: int = 10
                retry_config: RetryConfiguration = RetryConfiguration(
                    exponent=exponent, delay=delay, max_tries=max_tries
                )

                # WHEN / THEN
                with raises(TypeError):
                    # noinspection PyTypeChecker
                    retry_config.get_waiting_interval(None)

            def test_zero_retry_number__should__raise_value_error(self) -> None:
                """When retry_number is zero, should raise a ValueError exception"""
                # GIVEN
                exponent: int = 7
                delay: float = 0.5
                max_tries: int = 10
                retry_config: RetryConfiguration = RetryConfiguration(
                    exponent=exponent, delay=delay, max_tries=max_tries
                )

                # WHEN / THEN
                with raises(ValueError) as error:
                    # noinspection PyTypeChecker
                    retry_config.get_waiting_interval(0)

                assert str(error.value) == "retry_number must be greater than 0"

            def test_negative_retry_number__should__raise_value_error(self) -> None:
                """When retry_number is negative, should raise a ValueError exception"""
                # GIVEN
                exponent: int = 7
                delay: float = 0.5
                max_tries: int = 10
                retry_config: RetryConfiguration = RetryConfiguration(
                    exponent=exponent, delay=delay, max_tries=max_tries
                )

                # WHEN / THEN
                with raises(ValueError) as error:
                    # noinspection PyTypeChecker
                    retry_config.get_waiting_interval(-1)

                assert str(error.value) == "retry_number must be greater than 0"

            def test_retry_number_greater_than_max_tries__should__raise_value_error(self) -> None:
                """When retry_number is greater than max_tries, should raise a ValueError exception"""
                # GIVEN
                exponent: int = 7
                delay: float = 0.5
                max_tries: int = 10
                retry_config: RetryConfiguration = RetryConfiguration(
                    exponent=exponent, delay=delay, max_tries=max_tries
                )

                # WHEN / THEN
                with raises(ValueError) as error:
                    # noinspection PyTypeChecker
                    retry_config.get_waiting_interval(max_tries + 1)

                assert str(error.value) == "retry_number greater than configured max_tries"


class TestToRetryable:
    """to_retryable function tests"""
    _EXPONENT: int = 2
    _DELAY: float = 0.01
    _MAX_TRIES: int = 3
    _RETRY_CONFIG: RetryConfiguration = RetryConfiguration(exponent=_EXPONENT, delay=_DELAY, max_tries=_MAX_TRIES)
    _SEMI_FAILURE_CALL: int = 0

    @staticmethod
    def _successful_function(i: int) -> int:
        print(i)
        return i * i

    @staticmethod
    def _failure_function(i: int) -> int:
        print(i)
        raise ValueError(i)

    @staticmethod
    def _semi_failure_function(i: int) -> int:
        TestToRetryable._SEMI_FAILURE_CALL = (TestToRetryable._SEMI_FAILURE_CALL + 1) % 2
        return TestToRetryable._failure_function(i) \
            if TestToRetryable._SEMI_FAILURE_CALL == 1 \
            else TestToRetryable._successful_function(i)

    class TestNominalCase:

        @patch('builtins.print')
        def test_callable_in_success__should__return_retryable_callable_with_success_value(
                self, counter_mock: MagicMock
        ) -> None:
            """When the given callable always succeed, should return a retryable one and execution should succeed"""
            # GIVEN TestToRetryable.RETRY_CONFIG
            test_callable: Function[int, int] = TestToRetryable._successful_function

            # WHEN
            retryable_callable: Function[int, int] = retry.to_retryable(test_callable)

            # THEN
            param: int = 2
            assert retryable_callable(param) == (param * param)
            counter_mock.assert_called_once_with(param)
            assert counter_mock.call_count == 1

        @patch('builtins.print')
        def test_failure_callable_with_default_on_failure__should__return_retryable_and_raise_error_after_retry(
                self, counter_mock: MagicMock
        ) -> None:
            """When the given callable always failed, should return a retryable which raises an error on execution"""
            # GIVEN TestToRetryable.RETRY_CONFIG
            test_callable: Function[int, int] = TestToRetryable._failure_function

            # WHEN
            retryable_callable: Function[int, int] = \
                retry.to_retryable(test_callable, retry_config=TestToRetryable._RETRY_CONFIG)

            # THEN
            param: int = 2
            expected_min_duration: float = sum(
                [
                    TestToRetryable._RETRY_CONFIG.get_waiting_interval(n_retry)
                    for n_retry in range(1, TestToRetryable._MAX_TRIES)
                ]
            )
            exec_duration: float = time()
            with raises(ValueError) as error:
                retryable_callable(param)
            exec_duration = time() - exec_duration
            assert exec_duration >= expected_min_duration
            counter_mock.assert_called_with(param)
            assert counter_mock.call_count == TestToRetryable._MAX_TRIES
            assert str(error.value) == str(param)

        @patch('builtins.print')
        def test_failure_callable_with_on_failure__should__return_retryable_and_call_on_failure_parameter(
                self, counter_mock: MagicMock
        ) -> None:
            """
            When the given callable always failed and on_failure is provided, should return a retryable which calls the
            given on_failure function
            """
            # GIVEN TestToRetryable.RETRY_CONFIG
            test_callable: Function[int, int] = TestToRetryable._failure_function
            default_value: int = 7
            on_failure: Function[BaseException, int] = lambda _: default_value

            # WHEN
            retryable_callable: Function[int, int] = \
                retry.to_retryable(test_callable, retry_config=TestToRetryable._RETRY_CONFIG, on_failure=on_failure)

            # THEN
            param: int = 2
            expected_min_duration: float = sum(
                [
                    TestToRetryable._RETRY_CONFIG.get_waiting_interval(n_retry)
                    for n_retry in range(1, TestToRetryable._MAX_TRIES)
                ]
            )
            exec_duration: float = time()
            assert retryable_callable(param) == default_value
            exec_duration = time() - exec_duration
            counter_mock.assert_called_with(param)
            assert counter_mock.call_count == TestToRetryable._MAX_TRIES
            assert exec_duration >= expected_min_duration

        @patch('builtins.print')
        def test_semi_failure_callable__should__raise_callable_error_after_retry(
                self, counter_mock: MagicMock
        ) -> None:
            """
            When the given callable failed once, should return a retryable and execution should succeed after one retry
            """
            # GIVEN TestToRetryable.RETRY_CONFIG
            TestToRetryable._SEMI_FAILURE_CALL = 0
            test_callable: Function[int, int] = TestToRetryable._semi_failure_function

            # WHEN
            retryable_callable: Function[int, int] = \
                retry.to_retryable(test_callable, retry_config=TestToRetryable._RETRY_CONFIG)

            # THEN
            param: int = 2
            expected_min_duration: float = TestToRetryable._RETRY_CONFIG.get_waiting_interval(1)
            exec_duration: float = time()
            assert retryable_callable(param) == (param * param)
            exec_duration = time() - exec_duration
            assert exec_duration >= expected_min_duration
            counter_mock.assert_called_with(param)
            assert counter_mock.call_count == 2

        class TestErrorCase:
            def test_none_sync_callable__should__raise_type_error(self) -> None:
                """When sync_callable is None, should raise a TypeError exception"""
                # GIVEN / WHEN / THEN
                with raises(TypeError):
                    # noinspection PyTypeChecker
                    retry.to_retryable(sync_callable=None)

            def test_none_retry_config__should__raise_type_error(self) -> None:
                """When retry_config is None, should raise a TypeError exception"""
                # GIVEN
                test_callable: Function[int, int] = TestToRetryable._semi_failure_function

                # WHEN / THEN
                with raises(TypeError):
                    # noinspection PyTypeChecker
                    retry.to_retryable(sync_callable=test_callable, retry_config=None)

            def test_none_on_failure__should__raise_type_error(self) -> None:
                """When on_failure is None, should raise a TypeError exception"""
                # GIVEN
                test_callable: Function[int, int] = TestToRetryable._semi_failure_function

                # WHEN / THEN
                with raises(TypeError):
                    # noinspection PyTypeChecker
                    retry.to_retryable(sync_callable=test_callable, on_failure=None)


class TestToAsyncRetryable:
    """to_async_retryable function tests"""
    _EXPONENT: int = 2
    _DELAY: float = 0.01
    _MAX_TRIES: int = 3
    _RETRY_CONFIG: RetryConfiguration = RetryConfiguration(exponent=_EXPONENT, delay=_DELAY, max_tries=_MAX_TRIES)
    _SEMI_FAILURE_CALL: int = 0

    @staticmethod
    async def _successful_function(i: int) -> int:
        print(i)
        return i * i

    @staticmethod
    async def _failure_function(i: int) -> int:
        print(i)
        raise ValueError(i)

    @staticmethod
    async def _semi_failure_function(i: int) -> int:
        TestToAsyncRetryable._SEMI_FAILURE_CALL = (TestToAsyncRetryable._SEMI_FAILURE_CALL + 1) % 2
        return await TestToAsyncRetryable._failure_function(i) \
            if TestToAsyncRetryable._SEMI_FAILURE_CALL == 1 \
            else await TestToAsyncRetryable._successful_function(i)

    class TestNominalCase:

        @patch('builtins.print')
        async def test_callable_in_success__should__return_retryable_callable_with_success_value(
                self, counter_mock: MagicMock
        ) -> None:
            """When the given callable always succeed, should return a retryable one and execution should succeed"""
            # GIVEN TestToAsyncRetryable.RETRY_CONFIG
            test_callable: Function[int, Awaitable[int]] = TestToAsyncRetryable._successful_function

            # WHEN
            retryable_callable: Function[int, Awaitable[int]] = retry.to_async_retryable(test_callable)

            # THEN
            param: int = 2
            assert await retryable_callable(param) == (param * param)
            counter_mock.assert_called_once_with(param)
            assert counter_mock.call_count == 1

        @patch('builtins.print')
        async def test_failure_callable_with_default_on_failure__should__return_retryable_and_raise_error_after_retry(
                self, counter_mock: MagicMock
        ) -> None:
            """When the given callable always failed, should return a retryable which raises an error on execution"""
            # GIVEN TestToAsyncRetryable.RETRY_CONFIG
            test_callable: Function[int, Awaitable[int]] = TestToAsyncRetryable._failure_function

            # WHEN
            retryable_callable: Function[int, Awaitable[int]] = \
                retry.to_async_retryable(test_callable, retry_config=TestToAsyncRetryable._RETRY_CONFIG)

            # THEN
            param: int = 2
            expected_min_duration: float = sum(
                [
                    TestToAsyncRetryable._RETRY_CONFIG.get_waiting_interval(n_retry)
                    for n_retry in range(1, TestToAsyncRetryable._MAX_TRIES)
                ]
            )
            exec_duration: float = time()
            with raises(ValueError) as error:
                await retryable_callable(param)
            exec_duration = time() - exec_duration
            assert exec_duration >= expected_min_duration
            counter_mock.assert_called_with(param)
            assert counter_mock.call_count == TestToAsyncRetryable._MAX_TRIES
            assert str(error.value) == str(param)

        @patch('builtins.print')
        async def test_failure_callable_with_on_failure__should__return_retryable_and_call_on_failure_parameter(
                self, counter_mock: MagicMock
        ) -> None:
            """
            When the given callable always failed and on_failure is provided, should return a retryable which calls the
            given on_failure function
            """
            # GIVEN TestToAsyncRetryable.RETRY_CONFIG
            test_callable: Function[int, Awaitable[int]] = TestToAsyncRetryable._failure_function
            default_value: int = 7
            on_failure: Function[BaseException, int] = lambda _: default_value

            # WHEN
            retryable_callable: Function[int, Awaitable[int]] = \
                retry.to_async_retryable(test_callable, retry_config=TestToAsyncRetryable._RETRY_CONFIG,
                                         on_failure=on_failure)

            # THEN
            param: int = 2
            expected_min_duration: float = sum(
                [
                    TestToAsyncRetryable._RETRY_CONFIG.get_waiting_interval(n_retry)
                    for n_retry in range(1, TestToAsyncRetryable._MAX_TRIES)
                ]
            )
            exec_duration: float = time()
            assert await retryable_callable(param) == default_value
            exec_duration = time() - exec_duration
            counter_mock.assert_called_with(param)
            assert counter_mock.call_count == TestToAsyncRetryable._MAX_TRIES
            assert exec_duration >= expected_min_duration

        @patch('builtins.print')
        async def test_semi_failure_callable__should__raise_callable_error_after_retry(
                self, counter_mock: MagicMock
        ) -> None:
            """
            When the given callable failed once, should return a retryable and execution should succeed after one retry
            """
            # GIVEN TestToAsyncRetryable.RETRY_CONFIG
            TestToAsyncRetryable._SEMI_FAILURE_CALL = 0
            test_callable: Function[int, Awaitable[int]] = TestToAsyncRetryable._semi_failure_function

            # WHEN
            retryable_callable: Function[int, Awaitable[int]] = \
                retry.to_async_retryable(test_callable, retry_config=TestToAsyncRetryable._RETRY_CONFIG)

            # THEN
            param: int = 2
            expected_min_duration: float = TestToAsyncRetryable._RETRY_CONFIG.get_waiting_interval(1)
            exec_duration: float = time()
            assert await retryable_callable(param) == (param * param)
            exec_duration = time() - exec_duration
            assert exec_duration >= expected_min_duration
            counter_mock.assert_called_with(param)
            assert counter_mock.call_count == 2

        class TestErrorCase:
            def test_none_async_callable__should__raise_type_error(self) -> None:
                """When async_callable is None, should raise a TypeError exception"""
                # GIVEN / WHEN / THEN
                with raises(TypeError):
                    # noinspection PyTypeChecker
                    retry.to_async_retryable(async_callable=None)

            def test_none_retry_config__should__raise_type_error(self) -> None:
                """When retry_config is None, should raise a TypeError exception"""
                # GIVEN
                test_callable: Function[int, Awaitable[int]] = TestToAsyncRetryable._semi_failure_function

                # WHEN / THEN
                with raises(TypeError):
                    # noinspection PyTypeChecker
                    retry.to_async_retryable(async_callable=test_callable, retry_config=None)

            def test_none_on_failure__should__raise_type_error(self) -> None:
                """When on_failure is None, should raise a TypeError exception"""
                # GIVEN
                test_callable: Function[int, Awaitable[int]] = TestToAsyncRetryable._semi_failure_function

                # WHEN / THEN
                with raises(TypeError):
                    # noinspection PyTypeChecker
                    retry.to_async_retryable(async_callable=test_callable, on_failure=None)



"""tuples module tests"""

# Copyright 2022 the original author or authors (i.e.: St4rG00se for Dev4py).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from pytest import raises

from dev4py.utils import tuples


class TestEmptyTuple:
    """empty_tuple function tests"""

    class TestNominalCase:
        def test_should__return_an_empty_tuple(self) -> None:
            """Should return an empty tuple"""
            # GIVEN / WHEN
            result: tuple[int, ...] = tuples.empty_tuple()

            # THEN
            assert result == ()


class TestAppend:
    """append function tests"""

    class TestNominalCase:
        def test_existing_parameters__should__return_the_tuple_with_added_value(self) -> None:
            """When tuple and value are set, should return the tuple with added value"""
            # GIVEN
            tpl: tuple[int, ...] = (1, 2, 3)
            value: int = 4

            # WHEN
            result: tuple[int, ...] = tuples.append(tpl, value)

            # THEN
            assert result == (1, 2, 3, 4)
            assert tpl == (1, 2, 3)

        def test_none_value__should__return_the_tuple_with_added_value(self) -> None:
            """When tuple is set but value is none, should return the tuple with added value"""
            # GIVEN
            tpl: tuple[Optional[int], ...] = (1, 2, 3)
            value: Optional[int] = None

            # WHEN
            result: tuple[Optional[int], ...] = tuples.append(tpl, value)

            # THEN
            assert result == (1, 2, 3, None)
            assert tpl == (1, 2, 3)

    class TestErrorCase:
        def test_none_tuple__should__raise_type_error(self) -> None:
            """When the tuple is None, should raise TypeError exception"""
            # GIVEN
            value: int = 4

            # WHEN / THEN
            with raises(TypeError):
                # noinspection PyTypeChecker
                tuples.append(None, value)


class TestExtend:
    """extend function tests"""

    class TestNominalCase:
        def test_existing_parameters__should__add_elements_to_the_first_tuple_and_return_it(self) -> None:
            """When the tuples are set, should add all value from the second tuple to the first one and return it"""
            # GIVEN
            tpl_1: tuple[int, ...] = (1, 2, 3)
            tpl_2: tuple[int, ...] = (4, 5)

            # WHEN
            result: tuple[int, ...] = tuples.extend(tpl_1, tpl_2)

            # THEN
            assert result == (1, 2, 3, 4, 5)
            assert tpl_1 == (1, 2, 3)
            assert tpl_2 == (4, 5)

    class TestErrorCase:
        def test_none_first_tuple__should__raise_type_error(self) -> None:
            """When the first tuple is None, should raise TypeError exception"""
            # GIVEN
            tpl: tuple[int, ...] = tuples.empty_tuple()

            # WHEN / THEN
            with raises(TypeError):
                # noinspection PyTypeChecker
                tuples.extend(None, tpl)

        def test_none_second_tuple__should__raise_type_error(self) -> None:
            """When the second tuple is None, should raise TypeError exception"""
            # GIVEN
            tpl: tuple[int, ...] = tuples.empty_tuple()

            # WHEN / THEN
            with raises(TypeError):
                # noinspection PyTypeChecker
                tuples.extend(tpl, None)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2022 the original author or authors (i.e.: St4rG00se for Dev4py).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from io import TextIOWrapper
from os import environ as os_environ, pathsep as os_pathsep, getcwd as os_getcwd
from pathlib import Path
from subprocess import run as subprocess_run, CalledProcessError
from sys import argv as sys_argv, stderr as sys_stderr, stdin as sys_stdin, stdout as sys_stdout, exit as sys_exit
from textwrap import dedent
from typing import Final, TypeVar, Any

try:
    try:
        from tomli import loads as tomli_loads
    except ImportError:
        # noinspection PyProtectedMember
        # noinspection PyPackageRequirements
        # noinspection PyUnresolvedReferences
        from pip._vendor.tomli import loads as tomli_loads
except ImportError:
    print(
        "tomli is required (install pip or run this script from poetry venv after `poetry install --no-root` or "
        "`poetry update`)",
        file=sys_stderr)
    sys_exit(1)

# CONSTANTS
# - project
PROJECT_PATH: Final[Path] = Path(__file__).parent.absolute()

# - pyproject file consts
PROJECT_TOML_FILE_PATH: Final[str] = PROJECT_PATH.joinpath('pyproject.toml').as_posix()
PROJECT_SECTION: Final[str] = 'tool.poetry'
STRUCTURE_OPTION: Final[str] = 'packages'

# - sources / test default paths
DEFAULT_DIST_PATH: Final[str] = 'dist'
DEFAULT_BUILD_PATH: Final[str] = 'build'


# CLASSES
# - Project Configuration class
class ProjectProperties:
    T = TypeVar('T')

    def __init__(
            self,
            project_path: Path = PROJECT_PATH,
            toml_file_path: str = PROJECT_TOML_FILE_PATH,
            project_section: str = PROJECT_SECTION,
            structure_option: str = STRUCTURE_OPTION,
            dist_path: str = DEFAULT_DIST_PATH,
            build_path: str = DEFAULT_BUILD_PATH
    ):
        self.__toml_file_path: Final[str] = toml_file_path
        self.__project_section: Final[str] = project_section
        self.__structure_option: Final[str] = structure_option
        self.__toml_file_content: Final[dict[str, Any]] = self._load_toml()

        # - Global
        self.project_path: Final[Path] = project_path
        self.dist_path: Final[str] = dist_path
        self.build_path: Final[str] = build_path
        self.src_rsrc_paths: Final[list[str]] = self._get_sources_and_resources_paths()

    def _load_toml(self) -> dict[str, Any]:
        with open(self.__toml_file_path, "r", encoding="UTF-8") as toml_file:
            return tomli_loads(toml_file.read())

    def _get_option(self, section: str, option: str, default: T | None = None) -> T | None:
        option_path: list[str] = section.split('.')
        path_dict: Any = self.__toml_file_content
        for path in option_path:
            path_dict = path_dict.get(path)
            if not (path_dict and isinstance(path_dict, dict)):
                return default
        return path_dict.get(option, default)

    def _get_sources_and_resources_paths(self) -> list[str]:
        return [
            self.project_path.joinpath(pkg_cnf['from']).as_posix()
            for pkg_cnf in self._get_option(self.__project_section, self.__structure_option, default=[])
            if 'from' in pkg_cnf
        ]


# - Project commands abstract class
class ProjectCommand:
    """The ProjectCommand abstract class used to execute a new project command"""

    # noinspection PyMethodMayBeStatic
    # pylint: disable=no-self-use
    def get_command_cwd(self, properties: ProjectProperties) -> str:
        """
        Define the command cwd (by default it is the project path)
        :param properties: the project properties
        :return: the command line cwd
        """
        return properties.project_path.as_posix()

    def build_command_line(self, properties: ProjectProperties, args: list[str] | None = None) -> list[str]:
        """
        Build the command line to execute
        :param properties: the project properties
        :param args: the given argument list to use. Parse it with an ArgumentParser if necessary
        :return: the command argv list (ie sys.argv)
        """
        raise RuntimeError(f"abstract method -- subclass {self.__class__} must override")

    # noinspection PyMethodMayBeStatic
    # pylint: disable=no-self-use
    def finalize(self, properties: ProjectProperties) -> None:
        """
        Called once the command is executed in order to clean specific states
        :param properties: the project properties
        """


# -- Poetry Command (used as superclass)
class PoetryCommand(ProjectCommand):
    """Run poetry"""

    def build_command_line(self, properties: ProjectProperties, args: list[str] | None = None) -> list[str]:
        command_line: Final[list[str]] = ['poetry']
        if args:
            command_line.extend(args)
        return command_line


# -- Load dependencies Command
class LoadDepsCommand(PoetryCommand):
    """Install all dependencies (dev included)"""

    def build_command_line(self, properties: ProjectProperties, args: list[str] | None = None) -> list[str]:
        extended_args: Final[list[str]] = ['install', '--no-root']
        if args:
            extended_args.extend(args)
        return super().build_command_line(properties, extended_args)


# -- Clean Command
class CleanCommand(PoetryCommand):
    """Remove directories generated by the "build" commands (like 'sdist' or 'wheel')"""

    def build_command_line(self, properties: ProjectProperties, args: list[str] | None = None) -> list[str]:
        clean_script: Final[str] = dedent(f"""\
        from shutil import rmtree
        from pathlib import Path

        def rm_if_exists(path: Path) -> None:
            if path.is_dir():
                print(" |- Remove %s directory" % path)
                rmtree(path)
            elif path.is_file():
                print(" |- Remove %s file" % path)
                path.unlink()

        path_to_rm: list[Path] = [
            Path('{properties.dist_path}'),
            Path('{properties.build_path}'),
            Path('html/'),
            Path('.mutmut-cache')
        ]
        path_to_rm.extend(Path('.').rglob('.tox'))
        path_to_rm.extend(Path('.').rglob('.pytest_cache'))
        for path in path_to_rm:
            rm_if_exists(path)
        """)
        extended_args: Final[list[str]] = ['run', 'python', '-c', clean_script]
        if args:
            extended_args.extend(args)
        return super().build_command_line(properties, extended_args)


# -- Run Command
class RunCommand(PoetryCommand):
    """Run module which can be in the project structure without having to configure the PYTHONPATH"""

    def build_command_line(self, properties: ProjectProperties, args: list[str] | None = None) -> list[str]:
        # Use a subprocess in order to be in the good cwd in the poetry venv
        # It is useful when you call project.py from an outside directory
        run_script: Final[str] = dedent(f"""\
        from subprocess import run
        from sys import stdin, stdout, stderr, executable
        run(
            [executable, {str(args or '').strip('[]')}],
            stdin=stdin,
            stdout=stdout,
            stderr=stderr,
            check=True,
            cwd="{os_getcwd()}")
        """)
        extended_args: Final[list[str]] = ['run', 'python', '-c', run_script]
        if args:
            extended_args.extend(args)
        return super().build_command_line(properties, extended_args)


# -- Tox Command
class ToxCommand(PoetryCommand):
    """Run tox"""

    def build_command_line(self, properties: ProjectProperties, args: list[str] | None = None) -> list[str]:
        extended_args: Final[list[str]] = ['run', 'tox']
        if args:
            extended_args.extend(args)
        return super().build_command_line(properties, extended_args)


# -- Test Command
class TestCommand(ToxCommand):
    """Run configured unit tests"""
    SKIP_ENV_VAR: Final[str] = 'TOX_SKIP_ENV'

    def build_command_line(self, properties: ProjectProperties, args: list[str] | None = None) -> list[str]:
        os_environ[TestCommand.SKIP_ENV_VAR] = 'pylint|mutation|mypy'
        return super().build_command_line(properties, args)

    def finalize(self, properties: ProjectProperties) -> None:
        os_environ.pop(TestCommand.SKIP_ENV_VAR, None)


# -- Lint Command
class LintCommand(ToxCommand):
    """Run linter"""

    def build_command_line(self, properties: ProjectProperties, args: list[str] | None = None) -> list[str]:
        extended_args: Final[list[str]] = ['-e', 'pylint']
        if args:
            extended_args.extend(args)
        return super().build_command_line(properties, extended_args)


# -- Mutation Command
class MutationCommand(ToxCommand):
    """Run mutation tests"""

    def build_command_line(self, properties: ProjectProperties, args: list[str] | None = None) -> list[str]:
        extended_args: Final[list[str]] = ['-e', 'mutation']
        if args:
            extended_args.extend(args)
        return super().build_command_line(properties, extended_args)


# -- Typing check Command
class TypingCheckCommand(ToxCommand):
    """Run typing checker"""

    def build_command_line(self, properties: ProjectProperties, args: list[str] | None = None) -> list[str]:
        extended_args: Final[list[str]] = ['-e', 'mypy']
        if args:
            extended_args.extend(args)
        return super().build_command_line(properties, extended_args)


# -- Wheel Command
class WheelCommand(PoetryCommand):
    """Build Wheel archive"""

    def build_command_line(self, properties: ProjectProperties, args: list[str] | None = None) -> list[str]:
        extended_args: Final[list[str]] = ['build', '--format', 'wheel']
        if args:
            extended_args.extend(args)
        return super().build_command_line(properties, extended_args)


# -- Sdist Command
class SdistCommand(PoetryCommand):
    """Build sdist archive"""

    def build_command_line(self, properties: ProjectProperties, args: list[str] | None = None) -> list[str]:
        extended_args: Final[list[str]] = ['build', '--format', 'sdist']
        if args:
            extended_args.extend(args)
        return super().build_command_line(properties, extended_args)


# -- Upload Command
class UploadCommand(PoetryCommand):
    """Upload available deliveries"""

    def build_command_line(self, properties: ProjectProperties, args: list[str] | None = None) -> list[str]:
        extended_args: Final[list[str]] = ['publish']
        if args:
            extended_args.extend(args)
        return super().build_command_line(properties, extended_args)


# - Project command executor class
TCommandsRunner = TypeVar("TCommandsRunner", bound="CommandsRunner")


class CommandsRunner:
    def __init__(
            self,
            properties: ProjectProperties,
            stdin: TextIOWrapper = sys_stdin,
            stdout: TextIOWrapper = sys_stdout,
            stderr: TextIOWrapper = sys_stderr
    ):
        self.__project_properties: Final[ProjectProperties] = properties
        self.__stdin: Final[TextIOWrapper] = stdin
        self.__stdout: Final[TextIOWrapper] = stdout
        self.__stderr: Final[TextIOWrapper] = stderr
        self.__command_dict: Final[dict[str, ProjectCommand]] = {}

    def add_command(self, cmd_name: str, cmd: ProjectCommand) -> TCommandsRunner:
        self.__command_dict[cmd_name] = cmd
        return self

    def run(self) -> None:
        if sys_argv and len(sys_argv) > 1:
            argv: str = sys_argv[1]
            if argv in ("--help", "-h"):
                print(self._get_help_str(), file=self.__stdout)
                return

            current_cmd: ProjectCommand | None = self.__command_dict.get(argv)
            if not self.__command_dict.get(argv):
                print(f"Command unknown: '{argv}'", file=self.__stderr)
                return

            current_args: list[str] = []
            for argv in sys_argv[2:]:
                cmd: ProjectCommand | None = self.__command_dict.get(argv)
                if cmd:
                    self._run_process(current_cmd, current_args)
                    current_args.clear()
                    current_cmd = cmd
                else:
                    current_args.append(argv)
            self._run_process(current_cmd, current_args)
        else:
            print("Project command missing use --help or -h for help", file=self.__stderr)

    def _run_process(self, command: ProjectCommand, args: list[str]) -> None:
        try:
            subprocess_run(
                command.build_command_line(self.__project_properties, args),
                stdin=self.__stdin,
                stdout=self.__stdout,
                stderr=self.__stderr,
                check=True,
                cwd=command.get_command_cwd(self.__project_properties)
            )
        except CalledProcessError as e:
            print(f"Command error: [cmd: '{command.__class__.__name__}' | args: '{args}']", file=self.__stderr)
            sys_exit(e.returncode)
        finally:
            command.finalize(self.__project_properties)

    def _get_help_str(self):
        return "PROJECT COMMANDS WRAPPER:\n\n" \
               "Usage: python project.py <COMMAND_1> <arg1_1 ...> ... <COMMAND_N> <argN_1 ...>\n" \
               "\tNote: In order to get the wrapped command help, you can try python project.py <command> --help\n\n" \
               "Available commands are:\n" \
               + ''.join([f"  {cmd}   \t{cls.__doc__}\n" for cmd, cls in self.__command_dict.items()])


# FUNCTIONS
def run(properties: ProjectProperties) -> None:
    """"Execute given commands (from sys.argv) with the configured project structure in the PYTHONPATH"""
    # Configure CommandsRunner
    command_runner: Final[CommandsRunner] = CommandsRunner(properties=properties) \
        .add_command('load_deps', LoadDepsCommand()) \
        .add_command('clean', CleanCommand()) \
        .add_command('run', RunCommand()) \
        .add_command('tox', ToxCommand()) \
        .add_command('lint', LintCommand()) \
        .add_command('test', TestCommand()) \
        .add_command('typing', TypingCheckCommand()) \
        .add_command('mut', MutationCommand()) \
        .add_command('wheel', WheelCommand()) \
        .add_command('sdist', SdistCommand()) \
        .add_command('upload', UploadCommand())

    # Prepare PYTHONPATH
    project_paths: Final[list[str]] = properties.src_rsrc_paths
    pythonpath_env_var: Final[str] = 'PYTHONPATH'
    pythonpath: Final[str] = os_environ.get(pythonpath_env_var)
    if pythonpath:
        project_paths.append(pythonpath)

    # Run commands with project structure in the PYTHONPATH
    try:
        os_environ[pythonpath_env_var] = os_pathsep.join(project_paths)
        command_runner.run()
    finally:
        if pythonpath:
            os_environ[pythonpath_env_var] = pythonpath
        else:
            os_environ.pop(pythonpath_env_var, None)


# SHARED VARIABLES
project_properties = ProjectProperties()

# MAIN
if __name__ == '__main__':
    run(project_properties)



"""The `SimplePipeline` class is used in order to create a Pipeline with one operation per step"""

# Copyright 2022 the original author or authors (i.e.: St4rG00se for Dev4py).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from functools import partial
from typing import Generic, Final, cast

from dev4py.utils.objects import require_non_none
from dev4py.utils.types import OUT, IN, Function, R


class SimplePipeline(Generic[IN, OUT]):
    """A `SimplePipeline` with input of IN type and output of OUT type"""

    __CREATE_KEY: Final[object] = object()

    @classmethod
    def of(cls, handler: Function[IN, OUT]) -> SimplePipeline[IN, OUT]:
        """
        Returns a SimplePipeline[IN, OUT] initialized with the given handler

        Args:
            handler: The first pipeline handler

        Returns:
            SimplePipeline[IN, OUT]: The pipeline with the given handler as first operation

        """
        return SimplePipeline(handler, cls.__CREATE_KEY)

    def __init__(self, handler: Function[IN, OUT], create_key: object):
        """SimplePipeline private constructor: Constructs a Pipeline which processes IN to OUT value"""
        assert create_key == self.__CREATE_KEY, "SimplePipeline private constructor! Please use SimplePipeline.of"
        self._handler: Function[IN, OUT] = require_non_none(handler)

    def _add_handler_lambda(self, value: IN, handler: Function[OUT, R]) -> R:
        """
        private method to describe add_handler method new handler
        Note: lambda are not used in order to be compatible with multiprocessing (lambda are not serializable)
        """
        # lambda value: handler(self._handler(value))
        return handler(self._handler(value))

    def add_handler(self, handler: Function[OUT, R]) -> SimplePipeline[IN, R]:
        """
        Adds an operation/operation/step to the pipeline

        Args:
            handler: The handler to add

        Returns:
            SimplePipeline[IN, R]: The new pipeline where input is still of IN type but output is now of R type

        """
        require_non_none(handler)
        return SimplePipeline.of(cast(Function[IN, R], partial(self._add_handler_lambda, handler=handler)))

    def execute(self, value: IN) -> OUT:
        """
        Executes the current pipeline on the given value of IN type

        Args:
            value: The value of IN type

        Returns:
            OUT: the pipeline output of OUT type
        """
        return self._handler(value)



"""The `StepPipeline` class is used in order to create a Pipeline where each step can stop the execution"""

# Copyright 2022 the original author or authors (i.e.: St4rG00se for Dev4py).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from dataclasses import dataclass
from typing import Generic, Final, Any, Optional, Union

from dev4py.utils.JOptional import JOptional
from dev4py.utils.objects import require_non_none, require_non_none_else
from dev4py.utils.types import Function, T, N, IN, OUT


@dataclass
class StepResult(Generic[T]):
    """
    Represents a Step result

    Args:
        value: the result value of T type
        go_next: True if the next step must be executed, otherwise False
    """
    value: T
    go_next: bool = True


class _Step(Generic[IN, OUT]):

    def __init__(self, handler: Function[IN, StepResult[OUT]]):
        """_Step constructor"""
        self._handler: Final[Function[IN, StepResult[OUT]]] = require_non_none(handler)
        self._next: JOptional[_Step[OUT, Any]] = JOptional.empty()

    def add_next(self, next_handler: Function[OUT, StepResult[N]]) -> _Step[OUT, N]:
        """
        Adds a next step and returns it

        Args:
            next_handler: the next step handler

        Returns:
            _Step[OUT, N]: The new created step

        """
        next_step: _Step[OUT, N] = _Step(handler=require_non_none(next_handler))
        self._next = JOptional.of(next_step)
        return next_step

    def execute(self, value: IN) -> StepResult[Union[OUT, Any]]:
        """
        Executes the current step and if `go_next` is True call the next step if it exists. Returns the last executed
        step.

        Note: It means if the returned StepResult `go_next` field is TRUE all steps were executed. Otherwise, one step
        has stopped the execution.

        Args:
            value: The value of IN type

        Returns:
            StepResult[Union[OUT, Any]]: The last executed StepResult. Can be an intermediate step if `go_next` is False
        """
        step_result: Final[StepResult[OUT]] = self._handler(value)
        return self._next.get().execute(step_result.value) \
            if step_result.go_next and self._next.is_present() else step_result


class StepPipeline(Generic[IN, OUT]):
    __CREATE_KEY: Final[object] = object()

    @classmethod
    def of(cls, handler: Function[IN, StepResult[OUT]]) -> StepPipeline[IN, OUT]:
        """
        Returns a StepPipeline[IN, OUT] initialized with the given handler

        Note: the handler returned value is a `StepResult`. The function result value must be put in `value` field. To
        stop the current pipeline at this step `go_next` field must be set to False.

        Args:
            handler: the first step handler

        Returns:
            StepPipeline[IN, OUT]: The new pipeline StepPipeline which consumes an IN value and last step consumes an
            OUT value

        """
        return cls._of(step=_Step(handler))

    @classmethod
    def _of(cls, step: _Step[Any, OUT], head: Optional[_Step[IN, Any]] = None) -> StepPipeline[IN, OUT]:
        """
        !!! PRIVATE _OF CLASS METHOD !!!

        Returns a StepPipeline[IN, OUT] by using the given last step and the given head step

        Args:
            step: the last step
            head: the head step

        Returns:
            StepPipeline[IN, OUT]: The new pipeline StepPipeline which consumes an IN value and last step consumes an
            OUT value
        """
        return StepPipeline(cls.__CREATE_KEY, step=step, head=head)

    def __init__(self, create_key: object, step: _Step[Any, OUT], head: Optional[_Step[IN, Any]] = None):
        """
        StepPipeline private constructor: Constructs a Pipeline which consumes an IN value and last step consumes a OUT
        value
        """
        assert create_key == self.__CREATE_KEY, "StepPipeline private constructor! Please use StepPipeline.of"
        require_non_none(step)
        self._head: _Step[IN, Any] = require_non_none_else(head, step)
        self._last: _Step[Any, OUT] = step

    def add_handler(self, next_handler: Function[OUT, StepResult[N]]) -> StepPipeline[IN, N]:
        """
        Adds a new step to the pipeline by using the given handler and returns the new Pipeline

        Note: the handler returned value is a `StepResult`. The function result value must be put in `value` field. To
        stop the current pipeline at this step `go_next` field must be set to False.

        Args:
            next_handler: the new step handler

        Returns:
            StepPipeline[IN, N]: The new pipeline StepPipeline which consumes an IN value and last step consumes an N
                value

        """
        return StepPipeline._of(step=self._last.add_next(next_handler), head=self._head)

    def execute(self, value: IN) -> StepResult[Union[OUT, Any]]:
        """
        Executes the current pipeline and returns the last executed Step

        Note: The last executed step can be an intermediate one if `go_next` is false. If `go_next` is True it means
        that the pipeline was totally executed for the given value.

        Args:
            value: The value of IN type

        Returns:
            StepResult[Union[OUT, Any]]: The last executed StepResult. Can be an intermediate step if `go_next` is False

        """
        return self._head.execute(value)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src/test/python'}

packages = \
['python',
 'python.dev4py.utils',
 'python.dev4py.utils.pipeline',
 'python.dev4py.utils_test',
 'python.dev4py.utils_test.pipeline']

package_data = \
{'': ['*']}

modules = \
['project']
setup_kwargs = {
    'name': 'dev4py-utils',
    'version': '3.4.0',
    'description': 'A set of Python regularly used classes/functions',
    'long_description': '# Dev4py-utils\n\nA set of Python regularly used classes/functions\n\n[![ci](https://github.com/dev4py/dev4py-utils/actions/workflows/ci.yml/badge.svg?event=push&branch=main)](https://github.com/dev4py/dev4py-utils/actions/workflows/ci.yml) \\\n[![Last release](https://github.com/dev4py/dev4py-utils/actions/workflows/on_release.yml/badge.svg)](https://github.com/dev4py/dev4py-utils/actions/workflows/on_release.yml) \\\n[![Weekly checks](https://github.com/dev4py/dev4py-utils/actions/workflows/weekly_checks.yml/badge.svg?branch=main)](https://github.com/dev4py/dev4py-utils/actions/workflows/weekly_checks.yml) \\\n[![Python >= 3.10.1](https://img.shields.io/badge/Python->=3.10.1-informational.svg?style=plastic&logo=python&logoColor=yellow)](https://www.python.org/) \\\n[![Maintainer](https://img.shields.io/badge/maintainer-St4rG00se-informational?style=plastic&logo=superuser)](https://github.com/St4rG00se) \\\n[![Maintenance](https://img.shields.io/badge/Maintained%3F-yes-green.svg?style=plastic&logo=github)](https://GitHub.com/Naereen/StrapDown.js/graphs/commit-activity) \\\n[![License: Apache-2.0](https://img.shields.io/badge/License-Apache_2.0-yellow.svg?style=plastic&logo=github)](https://opensource.org/licenses/Apache-2.0)\n\n## Table of contents\n\n- [Quickstart](#quickstart)\n- [Project template](#project-template)\n- [Project links](#project-links)\n- [Dev4py-utils modules](#dev4py-utils-modules)\n  - [dev4py.utils.AsyncJOptional](#dev4pyutilsasyncjoptional)\n  - [dev4py.utils.awaitables](#dev4pyutilsawaitables)\n  - [dev4py.utils.collectors](#dev4pyutilscollectors)\n  - [dev4py.utils.dicts](#dev4pyutilsdicts)\n  - [dev4py.utils.iterables](#dev4pyutilsiterables)\n  - [dev4py.utils.JOptional](#dev4pyutilsjoptional)\n  - [dev4py.utils.lists](#dev4pyutilslists)\n  - [dev4py.utils.objects](#dev4pyutilsobjects)\n  - [dev4py.utils.pipeline](#dev4pyutilspipeline)\n  - [dev4py.utils.retry](#dev4pyutilsretry)\n  - [dev4py.utils.Stream](#dev4pyutilsstream)\n  - [dev4py.utils.tuples](#dev4pyutilstuples)\n  - [dev4py.utils.types](#dev4pyutilstypes)\n\n## Quickstart\n\n```shell\npip install dev4py-utils\n```\n\n## Project template\n\nThis project is based on [pymsdl_template](https://github.com/dev4py/pymsdl_template)\n\n## Project links\n\n- [Documentation](https://htmlpreview.github.io/?https://github.com/dev4py/dev4py-utils/blob/main/docs/dev4py/utils.html)\n- [PyPi project](https://pypi.org/project/dev4py-utils/)\n\n## Dev4py-utils modules\n\n### dev4py.utils.AsyncJOptional\n\n[AsyncJOptional documentation](https://htmlpreview.github.io/?https://github.com/dev4py/dev4py-utils/blob/main/docs/dev4py/utils/AsyncJOptional.html)\n\n> ***Note:** [AsyncJOptional](src/main/python/dev4py/utils/AsyncJOptional.py) class is designed in order to simplify\n> JOptional with async mapper.*\n\n> ***Note:** AsyncJOptional support T or Awaitable[T] values. That\'s why some checks are done when terminal operation is\n> called with `await`.*\n\nExamples:\n\n```python\nimport asyncio\nfrom dev4py.utils import AsyncJOptional\n\ndef sync_mapper(i: int) -> int:\n  return i * 2\n\nasync def async_mapper(i: int) -> str:\n  return f"The value is {i}"\n\nasync def async_sample() -> None:\n  value: int = 1\n  await AsyncJOptional.of_noneable(value) \\\n    .map(sync_mapper) \\\n    .map(async_mapper) \\\n    .if_present(print)  # The value is 2\n\nasyncio.run(async_sample())\n```\n\n### dev4py.utils.awaitables\n\n[Awaitables documentation](https://htmlpreview.github.io/?https://github.com/dev4py/dev4py-utils/blob/main/docs/dev4py/utils/awaitables.html)\n\n> ***Note:** [awaitables](src/main/python/dev4py/utils/awaitables.py) module provides a set of utility functions to\n> simplify Awaitable operations.*\n\nExamples:\n\n```python\nimport asyncio\nfrom dev4py.utils import awaitables, JOptional\n\n# is_awaitable sample\nawaitables.is_awaitable(asyncio.sleep(2))  # True\nawaitables.is_awaitable(print(\'Hello\'))  # False\n\n\n# to_sync_or_async_param_function sample\ndef mapper(s: str) -> str:\n    return s + \'_suffix\'\n\nasync def async_mapper(s: str) -> str:\n    await asyncio.sleep(1)\n    return s + \'_async_suffix\'\n\nasync def async_test():\n    # Note: mapper parameter is str and async_mapper returns an Awaitable[str] so we have to manage it\n    # Note: !WARNING! Since 3.0.0 see AsyncJOptional / JOptional to_async_joptional method\n    result: str = await JOptional.of("A value") \\\n      .map(async_mapper) \\\n      .map(awaitables.to_sync_or_async_param_function(mapper)) \\\n      .get()\n    print(result)  # A value_async_suffix_suffix\n\nasyncio.run(async_test())\n````\n\n### dev4py.utils.collectors\n\n[Collectors documentation](https://htmlpreview.github.io/?https://github.com/dev4py/dev4py-utils/blob/main/docs/dev4py/utils/collectors.html)\n\n> ***Note:** The [collectors](src/main/python/dev4py/utils/collectors.py) class is inspired by\n> [java.util.stream.Collectors](https://docs.oracle.com/en/java/javase/17/docs/api//java.base/java/util/stream/Collectors.html)*\n\nExamples:\n\n```python\nfrom dev4py.utils import Stream, collectors\n\nStream.of(\'a\', \'b\', \'c\').collect(collectors.to_list())  # [\'a\', \'b\', \'c\']\n```\n\n### dev4py.utils.dicts\n\n[Dicts documentation](https://htmlpreview.github.io/?https://github.com/dev4py/dev4py-utils/blob/main/docs/dev4py/utils/dicts.html)\n\n> ***Note:** [dicts](src/main/python/dev4py/utils/dicts.py) module provides a set of utility functions to\n> simplify dict operations.*\n\nExamples:\n\n```python\nfrom dev4py.utils import dicts\nfrom dev4py.utils.types import Supplier\n\n# is_dict sample\ndicts.is_dict("A str")  # False\ndicts.is_dict({\'key\': \'A dict value\'})  # True\n\n\n# get_value sample\nint_supplier: Supplier[int] = lambda: 3\ndictionary: dict[str, int] = {\'key_1\': 1, \'key_2\': 2}\n\ndicts.get_value(dictionary, \'key_1\')  # 1\ndicts.get_value(dictionary, \'key_3\')  # None\ndicts.get_value(dictionary, \'key_3\', int_supplier)  # 3\n\n\n# get_value_from_path sample\nstr_supplier: Supplier[str] = lambda: "a3"\ndeep_dictionary: dict[str, dict[int, str]] = { \\\n  \'a\': {1: \'a1\', 2: \'a2\'}, \\\n  \'b\': {1: \'b1\', 2: \'b2\'} \\\n}\n\ndicts.get_value_from_path(deep_dictionary, ["a", 1])  # \'a1\'\ndicts.get_value_from_path(deep_dictionary, ["c", 1])  # None\ndicts.get_value_from_path(deep_dictionary, ["a", 3])  # None\ndicts.get_value_from_path(deep_dictionary, ["a", 3], str_supplier)  # \'a3\'\n````\n\n### dev4py.utils.iterables\n\n[Iterables documentation](https://htmlpreview.github.io/?https://github.com/dev4py/dev4py-utils/blob/main/docs/dev4py/utils/iterables.html)\n\n> ***Note:** The [iterables](src/main/python/dev4py/utils/iterables.py) module provides a set of utility functions to simplify\n> iterables operations.*\n\nExample:\n\n```python\nfrom typing import Iterator\n\nfrom dev4py.utils import iterables\n\nvalues: range = range(0, 10)\nchunks: Iterator[list[int]] = iterables.get_chunks(values, 3)\n[chunk for chunk in chunks]  # [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9]]\n```\n\n### dev4py.utils.JOptional\n\n[JOptional documentation](https://htmlpreview.github.io/?https://github.com/dev4py/dev4py-utils/blob/main/docs/dev4py/utils/JOptional.html)\n\n> ***Note:** [JOptional](src/main/python/dev4py/utils/JOptional.py) class is inspired by\n> [java.util.Optional](https://docs.oracle.com/en/java/javase/17/docs/api//java.base/java/util/Optional.html)\n> class with some adds (like `peek` method).*\n\nExamples:\n\n```python\nfrom dev4py.utils import JOptional\n\nvalue: int = 1\nJOptional.of_noneable(value) \\\n  .map(lambda v: f"The value is {v}") \\\n  .if_present(print)  # The value is 1\n```\n\n### dev4py.utils.lists\n\n[Lists documentation](https://htmlpreview.github.io/?https://github.com/dev4py/dev4py-utils/blob/main/docs/dev4py/utils/lists.html)\n\n> ***Note:** [lists](src/main/python/dev4py/utils/lists.py) module provides a set of utility functions to simplify lists\n> operations.*\n\nExamples:\n\n```python\nfrom dev4py.utils import lists\n\n# empty sample\nlst: list[int] = lists.empty_list()  # []\n\n# append sample\nlst: list[int] = [1, 2, 3, 4]\napp_lst: list[int] = lists.append(lst, 5)  # [1, 2, 3, 4, 5]\n# - Note: lst == app_lst\n\n# extend sample\nlst: list[int] = [1, 2, 3, 4]\nlst2: list[int] = [5, 6, 7, 8]\next_lst: list[int] = lists.extend(lst, lst2)  # [1, 2, 3, 4, 5, 6, 7, 8]\n# - Note: lst == ext_lst\n```\n\n### dev4py.utils.objects\n\n[Objects documentation](https://htmlpreview.github.io/?https://github.com/dev4py/dev4py-utils/blob/main/docs/dev4py/utils/objects.html)\n\n> ***Note:** The [objects](src/main/python/dev4py/utils/objects.py) module is inspired by\n> [java.util.Objects](https://docs.oracle.com/en/java/javase/17/docs/api//java.base/java/util/Objects.html)\n> class.*\n\nExamples:\n\n```python\nfrom dev4py.utils import objects\n\n# non_none sample\nvalue = None\nobjects.non_none(value)  # False\n\n# require_non_none sample\nvalue = "A value"\nobjects.require_non_none(value)  # \'A value\'\n\n# to_string sample\nvalue = None\ndefault_value: str = "A default value"\nobjects.to_string(value, default_value)  # \'A default value\'\n```\n\n### dev4py.utils.pipeline\n\n[Pipeline documentation](https://htmlpreview.github.io/?https://github.com/dev4py/dev4py-utils/blob/main/docs/dev4py/utils/pipeline.html)\n\n> ***Note:** The [pipeline](src/main/python/dev4py/utils/pipeline) package provides a set of Pipeline class describing\n> different kind of pipelines.*\n\nExamples:\n\n```python\nfrom dev4py.utils.pipeline import SimplePipeline, StepPipeline, StepResult\n\n# SimplePipeline sample\npipeline: SimplePipeline[int, str] = SimplePipeline.of(lambda i: i * i) \\\n    .add_handler(str) \\\n    .add_handler(lambda s: f"Result: {s} | Type: {type(s)}")\n\npipeline.execute(10)  # "Result: 100 | Type: <class \'str\'>"\n\n\n# StepPipeline sample\n# Note: StepPipeline can be stopped at each step by setting `go_next` value to False\npipeline: StepPipeline[int, str] = StepPipeline.of(lambda i: StepResult(i * i)) \\\n    .add_handler(lambda i: StepResult(value=str(i), go_next=i < 150)) \\\n    .add_handler(lambda s: StepResult(f"Result: {s} | Type: {type(s)}"))\n\npipeline.execute(10)  # StepResult(value="Result: 100 | Type: <class \'str\'>", go_next=True)\n# - Note: When the pipeline is fully completed, `go_next` is True\npipeline.execute(15)  # StepResult(value=\'225\', go_next=False)\n# - Note: Even if the pipeline is not fully completed, the last StepResult is returned with `go_next=False`\n```\n\n### dev4py.utils.retry\n\n[Retry documentation](https://htmlpreview.github.io/?https://github.com/dev4py/dev4py-utils/blob/main/docs/dev4py/utils/retry.html)\n\n> ***Note:** The [retry](src/main/python/dev4py/utils/retry.py) module provides provides function to create retryable\n> callable from simple sync or async callables using exponential backoff*\n>\n> *Usage idea: network requests (HTTP, AMQP, MQTT, etc.) with retry on error*\n\nExamples:\n\n```python\nimport asyncio\nfrom time import time\nfrom typing import Awaitable\n\nfrom dev4py.utils.retry import RetryConfiguration, to_retryable, to_async_retryable\nfrom dev4py.utils.types import BiFunction\n\n# RetryConfiguration:\n# Note: exponential backoff used formula is \'delay * (exponent^retry_number)\'\n#\n#   => Example: For the following RetryConfiguration, waiting times in case of error are:\n#       * first try:                    0 sec (always 0 for the first try)\n#       * second try (/first retry):    1 sec (\'0.5 * (2^1)\')\n#       * third try (/second retry):    2 sec (\'0.5 * (2^2)\')\n#       * max_tries=3 => no fourth try (/third retry)\nretry_config: RetryConfiguration = RetryConfiguration(\n    delay=0.5,  # the exponential backoff delay in second (default: 0.1)\n    exponent=2,  # the exponential backoff exponent to determine delay between each try (default: 2)\n    max_tries=3  # max try number (first try included) (default: 3, i.e.: first try and 2 retry)\n)\n\n\n# to_retryable sample:\n# -> SUCCESSFUL CALL SAMPLE\ndef callable_sample(j: int, start_time: float) -> int:\n    print("callable_sample - call time: \'%.2f\'" % (time() - start_time))\n    return j ** 2\n\nretryable_sample: BiFunction[int, float, int] = to_retryable(sync_callable=callable_sample, retry_config=retry_config)\nresult: int = retryable_sample(3, time())  # result = 9\n# outputs:\n#  callable_sample - call time: \'0.00\'\n\n\n# -> IN ERROR CALL SAMPLE\ndef in_error_callable_sample(j: int, start_time: float) -> int:\n    print("in_error_callable_sample - call time: \'%.2f\'" % (time() - start_time))\n    raise ValueError(j)\n\nin_error_retryable_sample: BiFunction[int, float, int] = \\\n    to_retryable(sync_callable=in_error_callable_sample, retry_config=retry_config)\n# Note: By default the last raised exception is raised if max_tries is reach. You can change this behavior by setting\n#       the `on_failure` parameter\nresult: int = in_error_retryable_sample(3, time())\n# outputs:\n#  in_error_callable_sample - call time: \'0.00\'\n#  in_error_callable_sample - call time: \'1.00\'\n#  in_error_callable_sample - call time: \'3.00\'\n#  ValueError: 3\n\n\n# to_async_retryable sample:\nasync def in_error_async_callable_sample(j: int, start_time: float) -> int:\n    print("in_error_async_callable_sample - call time: \'%.2f\'" % (time() - start_time))\n    raise ValueError(j)\n\nasync def async_retryable_sample() -> None:\n    in_error_async_retryable_sample: BiFunction[int, float, Awaitable[int]] = \\\n        to_async_retryable(async_callable=in_error_async_callable_sample, retry_config=retry_config)\n    result: int = await in_error_async_retryable_sample(2, time())\n\nasyncio.run(async_retryable_sample())\n# outputs:\n#  in_error_async_callable_sample - call time: \'0.00\'\n#  in_error_async_callable_sample - call time: \'1.00\'\n#  in_error_async_callable_sample - call time: \'3.00\'\n#  ValueError: 2\n```\n\n### dev4py.utils.Stream\n\n[Stream documentation](https://htmlpreview.github.io/?https://github.com/dev4py/dev4py-utils/blob/main/docs/dev4py/utils/Stream.html)\n\n> ***Note:** [Stream](src/main/python/dev4py/utils/Stream.py) class is inspired by\n> [java.util.stream.Stream](https://docs.oracle.com/en/java/javase/17/docs/api//java.base/java/util/stream/Stream.html).*\n\nExamples:\n\n```python\nfrom concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor\nfrom time import sleep\n\nfrom dev4py.utils import Stream, ParallelConfiguration\n\n# Sequential sample\nStream.of(1, 2, 3, 4) \\\n    .map(str) \\\n    .peek(lambda s: sleep(0.5)) \\\n    .map(lambda s: f"Mapped value: {s}") \\\n    .to_list()  # [\'Mapped value: 1\', \'Mapped value: 2\', \'Mapped value: 3\', \'Mapped value: 4\']\n# - Note: Execution time around 2 sec due to the sleep call\n\n\n# Multithreading sample\nwith ThreadPoolExecutor(max_workers=2) as executor:\n    Stream.of(1, 2, 3, 4) \\\n        .parallel(parallel_config=ParallelConfiguration(executor=executor, chunksize=2)) \\\n        .map(str) \\\n        .peek(lambda s: sleep(0.5)) \\\n        .map(lambda s: f"Mapped value: {s}") \\\n        .to_list()  # [\'Mapped value: 3\', \'Mapped value: 4\', \'Mapped value: 1\', \'Mapped value: 2\']\n# - Note: Execution time around 1 sec due to the given ParallelConfiguration\n# - Note: Since this stream is (by default) unordered, results order is random\n\n\n# Multiprocessing sample\n# - Note: Due to use of Multiprocessing:\n#       * lambdas cannot be used since they cannot be pickled\n#       * This sample should be put in a python file in order to work\ndef _sleep(s: str) -> None:\n    # eq lambda s: sleep(0.5)\n    sleep(0.5)\n\ndef _mapper(s: str) -> str:\n    # eq lambda s: f"Mapped value: {s}"\n    return f"Mapped value: {s}"\n\nif __name__ == \'__main__\':\n    with ProcessPoolExecutor(max_workers=2) as executor:\n        Stream.of(1, 2, 3, 4) \\\n            .parallel(parallel_config=ParallelConfiguration(executor=executor, chunksize=2)) \\\n            .map(str) \\\n            .peek(_sleep) \\\n            .map(_mapper) \\\n            .to_list()\n\n# - Note: Execution time around 1 sec due to the given ParallelConfiguration\n#         (Reminder: Use Multiprocessing for CPU-bound tasks. In this case Multithreading is more appropriate)\n# - Note: Since this stream is (by default) unordered, results order is random\n```\n\n### dev4py.utils.tuples\n\n[Tuples documentation](https://htmlpreview.github.io/?https://github.com/dev4py/dev4py-utils/blob/main/docs/dev4py/utils/tuples.html)\n\n> ***Note:** [tuples](src/main/python/dev4py/utils/tuples.py) module provides a set of utility functions to simplify\n> tuples operations.*\n\nExamples:\n\n```python\nfrom dev4py.utils import tuples\n\n# empty sample\ntpl: tuple[int, ...] = tuples.empty_tuple()  # ()\n\n# append sample\ntpl: tuple[int, ...] = (1, 2, 3, 4)\napp_tpl: tuple[int, ...] = tuples.append(tpl, 5)  # (1, 2, 3, 4, 5)\n\n# extend sample\ntpl: tuple[int, ...] = (1, 2, 3, 4)\ntpl2: tuple[int, ...] = (5, 6, 7, 8)\next_tpl: tuple[int, ...] = tuples.extend(tpl, tpl2)  # (1, 2, 3, 4, 5, 6, 7, 8)\n```\n\n### dev4py.utils.types\n\n[Types documentation](https://htmlpreview.github.io/?https://github.com/dev4py/dev4py-utils/blob/main/docs/dev4py/utils/types.html)\n\n> ***Note:** The [types](src/main/python/dev4py/utils/types.py) module is inspired by\n> [java.util.function](https://docs.oracle.com/en/java/javase/17/docs/api//java.base/java/util/function/package-summary.html)\n> package.*\n\nExamples:\n\n```python\nfrom dev4py.utils.types import Function, Predicate, Consumer\n\n# Function sample\nint_to_str: Function[int, str] = lambda i: str(i)\nstr_result: str = int_to_str(1)  # \'1\'\n\n# Predicate sample\nstr_predicate: Predicate[str] = lambda s: s == "A value"\npred_result: bool = str_predicate("Value to test")  # False\n\n# Consumer sample\ndef sample(consumer: Consumer[str], value: str) -> None:\n    consumer(value)\n\ndef my_consumer(arg: str) -> None:\n    print(arg)\n\nsample(my_consumer, "My value")  # My value\n```\n',
    'author': 'St4rG00se',
    'author_email': 'st4rg00se@protonmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/dev4py/dev4py-utils',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'py_modules': modules,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



"""The `collectors` module provides a set of collectors inspired by java (java.util.stream.Collectors)"""

# Copyright 2022 the original author or authors (i.e.: St4rG00se for Dev4py).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from functools import partial
from typing import Generic, Any

from dev4py.utils import lists, dicts, objects, tuples
from dev4py.utils.objects import require_non_none
from dev4py.utils.types import BiConsumer, BiFunction, Supplier, T, R, Function, K, V


##############################
#  MODULE CLASSES/FUNCTIONS  #
##############################
@dataclass(frozen=True)
class Collector(Generic[T, R]):
    """
    Collector class: A generic dataclass to describe a collector[T,R]

    Args:
        supplier (Supplier[R]): A function that creates and returns a new result container
        accumulator (BiFunction[R, T, R]): A function that folds a value into a result container
        combiner (BiFunction[R, R, R]): A function that accepts two partial results and merges them
    """
    supplier: Supplier[R]
    accumulator: BiFunction[R, T, R]
    combiner: BiFunction[R, R, R]

    def __post_init__(self):
        """
        Post init values checker
        Raises:
            TypeError: Raises a TypeError at least one parameter is None
        """
        require_non_none(self.supplier)
        require_non_none(self.accumulator)
        require_non_none(self.combiner)


def of(supplier: Supplier[R], accumulator: BiFunction[R, T, R], combiner: BiFunction[R, R, R]) -> Collector[T, R]:
    """
    Returns a Collector by using the given parameters

    Args:
        supplier (Supplier[R]): A function that creates and returns a new result container
        accumulator (BiFunction[R, T, R]): A function that folds a value into a result container
        combiner (BiFunction[R, R, R]): A function that accepts two partial results and merges them

    Returns:
        collector: a collector[T,R] built with given parameters

    Raises:
        TypeError: Raises a TypeError at least one parameter is None
    """
    return Collector(
        supplier=require_non_none(supplier),
        accumulator=require_non_none(accumulator),
        combiner=require_non_none(combiner)
    )


def of_biconsumers(supplier: Supplier[R], accumulator: BiConsumer[R, T], combiner: BiConsumer[R, R]) -> Collector[T, R]:
    """
    Returns a Collector by using the given parameters (biconsumer instead of bifunction)

    Args:
        supplier (Supplier[R]): A function that creates and returns a new result container
        accumulator (BiConsumer[R, T]): A function that folds a value into a result container
        combiner (BiConsumer[R, R]): A function that accepts two partial results and merges them

    Returns:
        collector: a collector[T,R] built with given parameters

    Raises:
        TypeError: Raises a TypeError at least one parameter is None
    """
    return of(
        supplier=require_non_none(supplier),
        accumulator=_to_bifunction(accumulator),
        combiner=_to_bifunction(combiner)
    )


def to_list() -> Collector[T, list[T]]:
    """
    Returns a Collector that accumulates the input elements into a new list

    Returns:
        Collector[T, list[T]]: a Collector that accumulates the input elements into a new list
    """
    return of(
        supplier=lists.empty_list,
        accumulator=lists.append,
        combiner=lists.extend
    )


def to_dict(key_mapper: Function[T, K], value_mapper: Function[T, V]) -> Collector[T, dict[K, V]]:
    """
    Returns a Collector that accumulates elements into a dict whose keys and values are the result of applying the
    provided mapping functions to the input elements

    Args:
        key_mapper: a mapping function to produce keys
        value_mapper: a mapping function to produce values

    Returns:
         Collector[T, dict[K, V]]: a Collector which collects elements into a dict whose keys and values are the result
            of applying mapping functions to the input elements
    """
    require_non_none(key_mapper)
    require_non_none(value_mapper)
    return of_biconsumers(
        supplier=dicts.empty_dict,
        accumulator=partial(_to_dict_accumulator, key_mapper=key_mapper, value_mapper=value_mapper),
        combiner=_to_dict_combiner
    )


def to_none() -> Collector[T, None]:
    """
    Returns a Collector that always returns None

    Returns:
        Collector[T, None]: A collector that always returns None
    """
    return of(
        supplier=objects.to_none,
        accumulator=objects.to_none,
        combiner=objects.to_none
    )


def to_tuple() -> Collector[T, tuple[T, ...]]:
    """
    Returns a Collector that accumulates the input elements into a new tuple

    Returns:
        Collector[T, tuple[T, ...]]: a Collector that accumulates the input elements into a new tuple
    """
    return of(
        supplier=tuples.empty_tuple,
        accumulator=tuples.append,
        combiner=tuples.extend
    )


def to_counter() -> Collector[T, int]:
    """
    Returns a Collector accepting elements of type T that counts the number of input elements

    Returns:
        Collector[T, int]: a Collector accepting elements of type T that counts the number of input elements
    """
    return of(
        supplier=_to_counter_supplier,
        accumulator=_to_counter_accumulator,
        combiner=_to_counter_combiner
    )


##############################
#  PRIVATE MODULE FUNCTIONS  #
##############################
def _bifunction_from_biconsumer(t: T, r: R, biconsumer: BiConsumer[T, R]) -> T:
    """
    private function to create a BiFunction from a BiConsumer by using `partial` function
    Note: lambda or inner function are not used in order to be compatible with multiprocessing (lambda are not
    serializable)
    """
    require_non_none(biconsumer)(t, r)
    return t


def _to_bifunction(biconsumer: BiConsumer[T, R]) -> BiFunction[T, R, T]:
    """
    private function to create a BiFunction from a BiConsumer
    Note: lambda are not used in order to be compatible with multiprocessing (lambda are not serializable)
    """
    return partial(_bifunction_from_biconsumer, biconsumer=require_non_none(biconsumer))


def _to_dict_accumulator(d: dict[K, V], value: T, key_mapper: Function[T, K], value_mapper: Function[T, V]) -> None:
    """
    private function to represent a dict accumulator
    Note: lambda are not used in order to be compatible with multiprocessing (lambda are not serializable)
    """
    dicts.put_value(d, key_mapper(value), value_mapper(value))


def _to_dict_combiner(d1: dict[K, V], d2: dict[K, V]) -> None:
    """
    private function to represent a dict combiner
    Note: lambda are not used in order to be compatible with multiprocessing (lambda are not serializable)
    """
    dicts.update(d1, d2)


def _to_counter_supplier() -> int:
    """
    private function to represent a counter supplier
    Note: lambda are not used in order to be compatible with multiprocessing (lambda are not serializable)
    """
    return 0


def _to_counter_accumulator(i: int, value: Any) -> int:  # pylint: disable=W0613
    """
    private function to represent a counter accumulator
    Note: lambda are not used in order to be compatible with multiprocessing (lambda are not serializable)
    """
    return i + 1


def _to_counter_combiner(i1: int, i2: int) -> int:
    """
    private function to represent a counter combiner
    Note: lambda are not used in order to be compatible with multiprocessing (lambda are not serializable)
    """
    return i1 + i2



"""objects module tests"""

# Copyright 2022 the original author or authors (i.e.: St4rG00se for Dev4py).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Final

from pytest import raises

from dev4py.utils import objects
from dev4py.utils.types import Supplier


class TestIsNone:
    """is_none function tests"""

    class TestNominalCase:

        def test_none_value__should__return_true(self) -> None:
            """On None value should return True"""
            # GIVEN / WHEN
            result: bool = objects.is_none(None)

            # THEN
            assert result

        def test_non_none_value__should__return_false(self) -> None:
            """On non None value should return False"""
            # GIVEN / WHEN
            result: bool = objects.is_none("A non None value")

            # THEN
            assert not result

        def test_empty_value__should__return_false(self) -> None:
            """On empty value should return False"""
            # GIVEN / WHEN
            result: bool = objects.is_none({})

            # THEN
            assert not result

        def test_false_bool_value__should__return_false(self) -> None:
            """On false boolean value should return False"""
            # GIVEN / WHEN
            result: bool = objects.is_none(False)

            # THEN
            assert not result


class TestNonNone:
    """non_none function tests"""

    class TestNominalCase:

        def test_none_value__should__return_false(self) -> None:
            """On None value should return False"""
            # GIVEN / WHEN
            result: bool = objects.non_none(None)

            # THEN
            assert not result

        def test_non_none_value__should__return_true(self) -> None:
            """On non None value should return True"""
            # GIVEN / WHEN
            result: bool = objects.non_none("A non None value")

            # THEN
            assert result

        def test_empty_value__should__return_true(self) -> None:
            """On empty value should return True"""
            # GIVEN / WHEN
            result: bool = objects.non_none({})

            # THEN
            assert result

        def test_false_bool_value__should__return_true(self) -> None:
            """On false boolean value should return True"""
            # GIVEN / WHEN
            result: bool = objects.non_none(False)

            # THEN
            assert result


class TestRequireNonNone:
    """require_non_none function tests"""
    DEFAULT_ERROR_MESSAGE: Final[str] = "None object error"

    class TestNominalCase:

        def test_non_none_value__should__return_value(self) -> None:
            """When the value is not None, should return it"""
            # GIVEN
            value: str = "A value"

            # WHEN
            result: str = objects.require_non_none(value)

            # THEN
            assert result == value

        def test_none_value__should__raise_type_error(self) -> None:
            """When the value is None, should raise TypeError with default message"""
            # GIVEN / WHEN / THEN
            with raises(TypeError) as error:
                objects.require_non_none(None)

            assert str(error.value) == TestRequireNonNone.DEFAULT_ERROR_MESSAGE

        def test_none_value_with_message__should__raise_type_error_with_given_message(self) -> None:
            """When the value is None and a message is specified, should raise TypeError with given message"""
            # GIVEN
            message: str = "An error message"

            # WHEN / THEN
            with raises(TypeError) as error:
                objects.require_non_none(None, message=message)

            assert str(error.value) == message


class TestRequireNonNoneElse:
    """require_non_none_else function tests"""

    class TestNominalCase:
        def test_non_none_value__should__return_value(self) -> None:
            """When the value is not None, should return it"""
            # GIVEN
            value: str = "A value"
            default_value: str = "A default value"

            # WHEN
            result: str = objects.require_non_none_else(value, default=default_value)

            # THEN
            assert result == value

        def test_non_none_value_with_none_default__should__return_value(self) -> None:
            """When the value is not None and default value is None, should return the value"""
            # GIVEN
            value: str = "A value"
            # noinspection PyTypeChecker
            default_value: str = None

            # WHEN
            result: str = objects.require_non_none_else(value, default=default_value)

            # THEN
            assert result == value

        def test_none_value__should__return_the_default_value(self) -> None:
            """When the value is None, should return the default value"""
            # GIVEN
            value: Optional[str] = None
            default_value: str = "A default value"

            # WHEN
            result: str = objects.require_non_none_else(value, default=default_value)

            # THEN
            assert result == default_value

    class TestErrorCase:

        def test_none_value_with_none_default__should__raise_type_error(self) -> None:
            """When the value and default are None, should raise TypeError"""
            # GIVEN / WHEN / THEN
            with raises(TypeError) as error:
                objects.require_non_none_else(None, default=None)

            assert str(error.value) == TestRequireNonNone.DEFAULT_ERROR_MESSAGE


class TestRequireNonNoneElseGet:
    """require_non_none_else_get function tests"""

    class TestNominalCase:
        def test_non_none_value__should__return_value(self) -> None:
            """When the value is not None, should return it"""
            # GIVEN
            value: str = "A value"
            default_supplier: Supplier[str] = lambda: "A default value"

            # WHEN
            result: str = objects.require_non_none_else_get(value, supplier=default_supplier)

            # THEN
            assert result == value

        def test_non_none_value_with_none_default__should__return_value(self) -> None:
            """When the value is not None and default supplier value is None, should return the value"""
            # GIVEN
            value: str = "A value"
            # noinspection PyTypeChecker
            default_supplier: Supplier[str] = lambda: None

            # WHEN
            result: str = objects.require_non_none_else_get(value, supplier=default_supplier)

            # THEN
            assert result == value

        def test_non_none_value_with_none_supplier__should__return_value(self) -> None:
            """When the value is not None and default supplier is None, should return the value"""
            # GIVEN
            value: str = "A value"
            # noinspection PyTypeChecker
            default_supplier: Supplier[str] = None

            # WHEN
            result: str = objects.require_non_none_else_get(value, supplier=default_supplier)

            # THEN
            assert result == value

        def test_none_value__should__return_the_default_value(self) -> None:
            """When the value is None, should return the default value"""
            # GIVEN
            value: Optional[str] = None
            default_value: str = "A default value"
            default_supplier: Supplier[str] = lambda: default_value

            # WHEN
            result: str = objects.require_non_none_else_get(value, supplier=default_supplier)

            # THEN
            assert result == default_value

    class TestErrorCase:

        def test_none_value_with_none_default__should__raise_type_error(self) -> None:
            """When the value and default are None, should raise TypeError"""
            # GIVEN / WHEN / THEN
            with raises(TypeError) as error:
                objects.require_non_none_else_get(None, supplier=lambda: None)

            assert str(error.value) == TestRequireNonNone.DEFAULT_ERROR_MESSAGE

        def test_none_value_with_none_supplier__should__raise_type_error(self) -> None:
            """When the value and supplier are None, should raise TypeError"""
            # GIVEN / WHEN / THEN
            with raises(TypeError) as error:
                objects.require_non_none_else_get(None, None)

            assert str(error.value) == "Supplier cannot be None"


class TestToString:
    """to_string function tests"""

    class TestNominalCase:
        def test_non_none_value__should__return_value_as_str(self) -> None:
            """When the value is not None, should return it as str format"""
            # GIVEN
            value: int = 77

            # WHEN
            result: str = objects.to_string(value)

            # THEN
            assert result == "77"

        def test_non_none_value_with_non_none_default__should__return_value_as_str(self) -> None:
            """When the value and default value are not None, should return it as str format"""
            # GIVEN
            value: int = 77
            default_value: str = "A default value"

            # WHEN
            result: str = objects.to_string(value, default_value)

            # THEN
            assert result == "77"

        def test_none_value_with_non_none_default__should__return_default_value(self) -> None:
            """When the value is None and default value is not none, should return the default value"""
            # GIVEN
            default_value: str = "A default value"

            # WHEN
            result: str = objects.to_string(None, default_value)

            # THEN
            assert result == default_value

        def test_none_value_with_none_default__should__return_none_as_str(self) -> None:
            """When the value and default value are None, should return 'None' str"""
            # GIVEN / WHEN
            result: str = objects.to_string(None, None)

            # THEN
            assert result == str(None)


class TestAsyncRequireNonNone:
    """async_require_non_none function tests"""
    DEFAULT_ERROR_MESSAGE: Final[str] = "None async object error"

    class TestNominalCase:

        async def test_non_none_value__should__return_value(self) -> None:
            """When the value is not None, should return it"""
            # GIVEN
            value: str = "A value"

            # WHEN
            result: str = await objects.async_require_non_none(value)

            # THEN
            assert result == value

        async def test_non_none_async_value__should__return_value(self) -> None:
            """When the value an awaitable with non None result, should return it"""
            # GIVEN
            value: str = "A value"

            async def coroutine() -> str:
                return value

            # WHEN
            result: str = await objects.async_require_non_none(coroutine())

            # THEN
            assert result == value

        async def test_none_value__should__raise_type_error(self) -> None:
            """When the value is None, should raise TypeError with default message"""
            # GIVEN / WHEN / THEN
            with raises(TypeError) as error:
                await objects.async_require_non_none(None)

            assert str(error.value) == TestAsyncRequireNonNone.DEFAULT_ERROR_MESSAGE

        async def test_none_async_value__should__raise_type_error(self) -> None:
            """When the value is an awaitable with None result, should raise TypeError with default message"""

            # GIVEN
            async def coroutine() -> None:
                return None

            # WHEN / THEN
            with raises(TypeError) as error:
                await objects.async_require_non_none(coroutine())

            assert str(error.value) == TestAsyncRequireNonNone.DEFAULT_ERROR_MESSAGE

        async def test_none_value_with_message__should__raise_type_error_with_given_message(self) -> None:
            """When the value is None and a message is specified, should raise TypeError with given message"""
            # GIVEN
            message: str = "An error message"

            # WHEN / THEN
            with raises(TypeError) as error:
                await objects.async_require_non_none(None, message=message)

            assert str(error.value) == message


class TestAsyncRequireNonNoneElse:
    """async_require_non_none_else function tests"""

    class TestNominalCase:
        async def test_non_none_value__should__return_value(self) -> None:
            """When the value is not None, should return it"""
            # GIVEN
            value: str = "A value"
            default_value: str = "A default value"

            # WHEN
            result: str = await objects.async_require_non_none_else(value, default=default_value)

            # THEN
            assert result == value

        async def test_non_none_async_value__should__return_value(self) -> None:
            """When the value is a non None async value, should return it"""
            # GIVEN
            value: str = "A value"

            async def coroutine() -> str: return value

            default_value: str = "A default value"

            # WHEN
            result: str = await objects.async_require_non_none_else(coroutine(), default=default_value)

            # THEN
            assert result == value

        async def test_non_none_value_with_none_default__should__return_value(self) -> None:
            """When the value is not None and default value is None, should return the value"""
            # GIVEN
            value: str = "A value"
            # noinspection PyTypeChecker
            default_value: str = None

            # WHEN
            result: str = await objects.async_require_non_none_else(value, default=default_value)

            # THEN
            assert result == value

        async def test_non_none_async_value_with_none_default__should__return_value(self) -> None:
            """When the value is non None async value and default value is None, should return the value"""
            # GIVEN
            value: str = "A value"

            async def coroutine() -> str: return value

            # noinspection PyTypeChecker
            default_value: str = None

            # WHEN
            result: str = await objects.async_require_non_none_else(coroutine(), default=default_value)

            # THEN
            assert result == value

        async def test_none_value__should__return_the_default_value(self) -> None:
            """When the value is None, should return the default value"""
            # GIVEN
            value: Optional[str] = None
            default_value: str = "A default value"

            # WHEN
            result: str = await objects.async_require_non_none_else(value, default=default_value)

            # THEN
            assert result == default_value

        async def test_none_async_value__should__return_the_default_value(self) -> None:
            """When the value is None async value, should return the default value"""

            # GIVEN
            async def coroutine() -> Optional[str]: return None

            default_value: str = "A default value"

            # WHEN
            result: str = await objects.async_require_non_none_else(coroutine(), default=default_value)

            # THEN
            assert result == default_value

    class TestErrorCase:

        async def test_none_value_with_none_default__should__raise_type_error(self) -> None:
            """When the value and default are None, should raise TypeError"""
            # GIVEN / WHEN / THEN
            with raises(TypeError) as error:
                await objects.async_require_non_none_else(None, default=None)

            assert str(error.value) == TestRequireNonNone.DEFAULT_ERROR_MESSAGE

        async def test_none_async_value_with_none_default__should__raise_type_error(self) -> None:
            """When the async value and default are None, should raise TypeError"""

            # GIVEN
            async def coroutine() -> Optional[str]: return None

            # WHEN / THEN
            with raises(TypeError) as error:
                await objects.async_require_non_none_else(coroutine(), default=None)

            assert str(error.value) == TestRequireNonNone.DEFAULT_ERROR_MESSAGE


class TestAsyncRequireNonNoneElseGet:
    """async_require_non_none_else_get function tests"""

    class TestNominalCase:
        async def test_non_none_value__should__return_value(self) -> None:
            """When the value is not None, should return it"""
            # GIVEN
            value: str = "A value"
            default_supplier: Supplier[str] = lambda: "A default value"

            # WHEN
            result: str = await objects.async_require_non_none_else_get(value, supplier=default_supplier)

            # THEN
            assert result == value

        async def test_non_none_async_value__should__return_value(self) -> None:
            """When the value is non None async, should return it"""
            # GIVEN
            value: str = "A value"
            default_supplier: Supplier[str] = lambda: "A default value"

            async def coroutine() -> str: return value

            # WHEN
            result: str = await objects.async_require_non_none_else_get(coroutine(), supplier=default_supplier)

            # THEN
            assert result == value

        async def test_non_none_value_with_none_default__should__return_value(self) -> None:
            """When the value is not None and default supplier value is None, should return the value"""
            # GIVEN
            value: str = "A value"
            # noinspection PyTypeChecker
            default_supplier: Supplier[str] = lambda: None

            # WHEN
            result: str = await objects.async_require_non_none_else_get(value, supplier=default_supplier)

            # THEN
            assert result == value

        async def test_non_none_async_value_with_none_default__should__return_value(self) -> None:
            """When the value is non None async and default supplier value is None, should return the value"""
            # GIVEN
            value: str = "A value"
            # noinspection PyTypeChecker
            default_supplier: Supplier[str] = lambda: None

            async def coroutine() -> str: return value

            # WHEN
            result: str = await objects.async_require_non_none_else_get(coroutine(), supplier=default_supplier)

            # THEN
            assert result == value

        async def test_non_none_value_with_none_supplier__should__return_value(self) -> None:
            """When the value is not None and default supplier is None, should return the value"""
            # GIVEN
            value: str = "A value"
            # noinspection PyTypeChecker
            default_supplier: Supplier[str] = None

            # WHEN
            result: str = await objects.async_require_non_none_else_get(value, supplier=default_supplier)

            # THEN
            assert result == value

        async def test_non_none_async_value_with_none_supplier__should__return_value(self) -> None:
            """When the value is non None async and default supplier is None, should return the value"""
            # GIVEN
            value: str = "A value"
            # noinspection PyTypeChecker
            default_supplier: Supplier[str] = None

            async def coroutine() -> str: return value

            # WHEN
            result: str = await objects.async_require_non_none_else_get(coroutine(), supplier=default_supplier)

            # THEN
            assert result == value

        async def test_none_value__should__return_the_default_value(self) -> None:
            """When the value is None, should return the default value"""
            # GIVEN
            value: Optional[str] = None
            default_value: str = "A default value"
            default_supplier: Supplier[str] = lambda: default_value

            # WHEN
            result: str = await objects.async_require_non_none_else_get(value, supplier=default_supplier)

            # THEN
            assert result == default_value

        async def test_none_async_value__should__return_the_default_value(self) -> None:
            """When the value is async None, should return the default value"""
            # GIVEN
            default_value: str = "A default value"
            default_supplier: Supplier[str] = lambda: default_value

            async def coroutine() -> Optional[str]: return None

            # WHEN
            result: str = await objects.async_require_non_none_else_get(coroutine(), supplier=default_supplier)

            # THEN
            assert result == default_value

    class TestErrorCase:

        async def test_none_value_with_none_default__should__raise_type_error(self) -> None:
            """When the value and default are None, should raise TypeError"""
            # GIVEN / WHEN / THEN
            with raises(TypeError) as error:
                await objects.async_require_non_none_else_get(None, supplier=lambda: None)

            assert str(error.value) == TestRequireNonNone.DEFAULT_ERROR_MESSAGE

        async def test_none_async_value_with_none_default__should__raise_type_error(self) -> None:
            """When the value is async None and default are None, should raise TypeError"""

            # GIVEN
            async def coroutine() -> Optional[str]: return None

            # WHEN / THEN
            with raises(TypeError) as error:
                await objects.async_require_non_none_else_get(coroutine(), supplier=lambda: None)

            assert str(error.value) == TestRequireNonNone.DEFAULT_ERROR_MESSAGE

        async def test_none_value_with_none_supplier__should__raise_type_error(self) -> None:
            """When the value and supplier are None, should raise TypeError"""
            # GIVEN / WHEN / THEN
            with raises(TypeError) as error:
                await objects.async_require_non_none_else_get(None, None)

            assert str(error.value) == "Supplier cannot be None"

        async def test_none_async_value_with_none_supplier__should__raise_type_error(self) -> None:
            """When the value is None async and supplier are None, should raise TypeError"""

            # GIVEN
            async def coroutine() -> Optional[str]: return None

            # WHEN / THEN
            with raises(TypeError) as error:
                await objects.async_require_non_none_else_get(coroutine(), None)

            assert str(error.value) == "Supplier cannot be None"


class TestToNone:
    """to_none function tests"""

    class TestNominalCase:

        def test__should__return_none(self) -> None:
            """Should return None whatever the parameters"""
            # GIVEN / WHEN / THEN
            assert objects.to_none(1, 'a', named_param=True) is None


class TestToSelf:
    """to_self function tests"""

    class TestNominalCase:

        def test_existing_parameter__should__return_given_parameter(self) -> None:
            """Should return the given parameter"""
            # GIVEN
            obj: str = "A test str"

            # WHEN
            result: str = objects.to_self(obj)

            # THEN
            assert result == obj

        def test_none_parameter__should__return_none(self) -> None:
            """Should return the given parameter"""
            # GIVEN / WHEN
            result: None = objects.to_self(None)

            # THEN
            assert result is None
